import warnings

import matplotlib.pyplot as plt
import matplotlib

import pyDeltaRCM

# filter out the warning raised about no netcdf being found
warnings.filterwarnings("ignore", category=UserWarning)


n = 10
cm = matplotlib.cm.get_cmap('tab10')


# init delta model
with pyDeltaRCM.shared_tools._docs_temp_directory() as output_dir:
    delta = pyDeltaRCM.DeltaModel(
        out_dir=output_dir,
        resume_checkpoint='../../_resources/checkpoint')


_shp = delta.eta.shape


# manually call only the necessary paths
delta.init_water_iteration()
delta.run_water_iteration()

# lines cobbled together from compute_free_surface and finalize_free_surface
delta.sfc_visit, delta.sfc_sum = pyDeltaRCM.water_tools._accumulate_free_surface_walks(
    delta.free_surf_walk_inds, delta.free_surf_flag, delta.cell_type,
    delta.uw, delta.ux, delta.uy, delta.depth,
    delta._dx, delta._u0, delta.h0, delta._H_SL, delta._S0)
Hnew = delta.eta + delta.depth
Hnew[delta.sfc_visit > 0] = (delta.sfc_sum[delta.sfc_visit > 0] /
                             delta.sfc_visit[delta.sfc_visit > 0])
Hnew[Hnew < delta._H_SL] = delta._H_SL
Hnew[Hnew < delta.eta] = delta.eta[Hnew < delta.eta]

# now run the smoothing function to get grids
Hsmth = pyDeltaRCM.water_tools._smooth_free_surface(
    Hnew, delta.cell_type, delta._Nsmooth, delta._Csmooth)

# set up axis
fig, ax = plt.subplots(1, 3, sharex=True, sharey=True, figsize=(12, 4))

# fill in axis0
pyDeltaRCM.debug_tools.plot_domain(
    Hnew, ax=ax[0], grid=False, cmap='viridis')
ax[0].set_title('stage before smoothing')

# fill in axis1
pyDeltaRCM.debug_tools.plot_domain(
    Hsmth, ax=ax[1], grid=False, cmap='viridis')
ax[1].set_title('stage after smoothing')


# fill in axis2
_diff = Hsmth - Hnew
pyDeltaRCM.debug_tools.plot_domain(
    _diff, ax=ax[2], grid=False, cmap='RdBu',
    vmin=-0.1, vmax=0.1)
ax[2].set_title('difference (after - before)')


plt.tight_layout()
plt.show()



import matplotlib.pyplot as plt

import pyDeltaRCM

with pyDeltaRCM.shared_tools._docs_temp_directory() as output_dir:
    delta = pyDeltaRCM.DeltaModel(out_dir=output_dir)

    for _t in range(0, 5):
        delta.update()

    delta.finalize()


fig, ax = plt.subplots()
ax.imshow(delta.bed_elevation, vmax=-3)
plt.show()



#! /usr/bin/env python
import warnings
import logging
import datetime
import os

from .iteration_tools import iteration_tools
from .sed_tools import sed_tools
from .water_tools import water_tools
from .init_tools import init_tools
from .hook_tools import hook_tools
from .debug_tools import debug_tools
from .shared_tools import _get_version


class DeltaModel(iteration_tools, sed_tools, water_tools,
                 init_tools, hook_tools, debug_tools, object):
    """Main model class.

    Instantiating the model class is described in the :meth:`__init__` method
    below in detail, but generally model instantiation occurs via a model run
    YAML configuration file. These YAML configuration files define model
    parameters which are used in the run; read more about creating input YAML
    configuration files in the :doc:`/guides/user_guide`.

    Once the model has been instantiated, the model is updated via the
    :meth:`update`. This method coordinates the hydrology, sediment transport,
    subsidence, and stratigraphic operation that must occur during each
    timestep. The :meth:`update` should be called iteratively, to "run" the
    model.

    Finally, after all ``update`` steps are complete, the model should be
    finalized (:meth:`finalize`), such that files and data are appropriately
    closed and written to disk.
    """

    def __init__(self, input_file=None, defer_output=False, **kwargs):
        """Creates an instance of the pyDeltaRCM model.

        This method handles setting up the run, including parsing input files,
        initializing arrays, and initializing output routines.

        Parameters
        ----------
        input_file : `str`, `os.PathLike`, optional
            User model run configuration file.

        defer_output : `bool`, optional
            Whether to create the output netCDF file during initialization. In
            most cases, this can be ignored and left to default value `False`.
            However, for parallel simulations it may be necessary to defer the
            NetCDF file creation unitl the simualtion is assigned to the core
            it will compute on, so that the DeltaModel object remains
            pickle-able. Note, you will need to manually trigger the
            :obj:`init_output_file`, :obj:`output_data`, and
            :obj:`output_checkpoint` method if `defer_output` is `True`.

        **kwargs
            Optionally, any parameter typically specified in the YAML file can
            be passed as a keyword argument to the instantiation of the
            DeltaModel. Keywords will override the specification of any value
            in the YAML file. This functionality is intended for advanced use
            cases, and should not be preffered to specifying all inputs in a
            YAML configuration file.

        Returns
        -------

        Notes
        -----
        For more information regarding input configuration files, see the
        :doc:`/guides/user_guide`.

        """
        self.__pyDeltaRCM_version__ = _get_version()

        self._time = 0.
        self._time_iter = int(0)
        self._save_time_since_data = float("inf")  # force save on t==0
        self._save_iter = int(0)
        self._save_time_since_checkpoint = float("inf")  # force save on t==0

        self.input_file = input_file
        _src_dir = os.path.realpath(os.path.dirname(__file__))
        self.default_file = os.path.join(_src_dir, 'default.yml')

        # infrastructure for custom yaml parameters if needed
        if hasattr(self, 'subclass_parameters') is False:
            self.subclass_parameters = dict()  # init dict for custom params

        # check for any deprecated hooks
        self._check_deprecated_hooks()

        # import the input file
        self.hook_import_files()  # user hook
        self.import_files(kwargs)

        # initialize output folders and logger
        self.init_output_infrastructure()
        self.init_logger()

        # apply the configuration
        self.hook_process_input_to_model()
        self.process_input_to_model()

        # determine and set the model seed
        self.determine_random_seed()

        # create model variables based on configuration
        self.hook_create_other_variables()
        self.create_other_variables()

        # create the model domain based on configuration
        self.hook_create_domain()
        self.create_domain()
        self.hook_after_create_domain()

        # initialize the sediment router classes
        self.init_sediment_routers()

        # set up the subsidence fields
        self.init_subsidence()

        # if resume flag set to True, load checkpoint, open netCDF4
        if self.resume_checkpoint:
            # load values from the checkpoint and don't init final features
            self.hook_load_checkpoint()
            self.load_checkpoint(defer_output)

        else:
            # initialize the output file
            if not defer_output:
                self.hook_init_output_file()
                self.init_output_file()

                # record initial conditions
                self.hook_output_data()
                self.output_data()

                self.hook_output_checkpoint()
                self.output_checkpoint()

        _msg = 'Model initialization complete'
        self.log_info(_msg)
        self.log_model_time()

    def update(self):
        """Run the model for one full timestep.

        This method handles the input/output from the model, orchestrating the
        various morphodynamic and basin-scale processes (from sub-methods),
        and incrementing the model time-tracking attributes. This method
        calls, in sequence:

            * the routine to run one timestep (i.e., water surface estimation
              and sediment routing;
              :meth:`~pyDeltaRCM.iteration_tools.iteration_tools.solve_water_and_sediment_timestep`)
            * the basin subsidence update;
              (:meth:`~pyDeltaRCM.iteration_tools.iteration_tools.apply_subsidence`)
            * the timestep finalization routine (applying boundary conditions);
              (:meth:`~pyDeltaRCM.iteration_tools.iteration_tools.finalize_timestep`)
            * the internal time-tracking attributes of the model are updated
            * the routine to output data to the NetCDF file;
              (:meth:`~pyDeltaRCM.iteration_tools.iteration_tools.output_data`)
            * the routine to output a checkpoint file;
              (:meth:`~pyDeltaRCM.iteration_tools.iteration_tools.output_checkpoint`)

        If you attempt to override the ``update`` routine, you must implement
        these operations at a minimum. More likely, you can implement what you
        need to by utilizing a model `hook`, or overriding one of the methods
        called by `update`. For more information on customizing the model, see
        the :ref:`complete guide <customize_the_model>`.

        Parameters
        ----------

        Returns
        -------

        Raises
        ------
        RuntimeError
            If model has already been finalized via :meth:`finalize`.

        """
        if self._is_finalized:
            raise RuntimeError('Cannot update model, model already finalized!')

        # update the model, i.e., the actual model morphodynamics
        self.hook_solve_water_and_sediment_timestep()
        self.solve_water_and_sediment_timestep()

        self.hook_apply_subsidence()
        self.apply_subsidence()

        self.hook_finalize_timestep()
        self.finalize_timestep()
        self.hook_after_finalize_timestep()

        # update time-tracking fields
        self._time += self.dt
        self._save_time_since_data += self.dt
        self._save_time_since_checkpoint += self.dt
        self._time_iter += int(1)
        self.log_model_time()

        # record the state of the model if needed
        self.hook_output_data()
        self.output_data()

        # save a checkpoint if needed
        self.hook_output_checkpoint()
        self.output_checkpoint()

    def finalize(self):
        """Finalize the model run.

        Finalization includes saving output stratigraphy, closing relevant
        files on disk and clearing variables.

        Parameters
        ----------

        Returns
        -------

        """
        _msg = 'Finalize the model run'
        self.log_info(_msg)

        if self._is_finalized:
            raise RuntimeError('Cannot finalize model, '
                               'model already finalized!')

        try:
            self.output_netcdf.close()
            _msg = 'Closed output NetCDF4 file'
            self.log_info(_msg, verbosity=1)
        except AttributeError:
            self.log_info('No output NetCDF4 file to close.')
        except Exception as e:
            self.logger.error('Failed to close output NetCDF4 file')
            self.logger.exception(e)

        self._is_finalized = True

    @property
    def out_dir(self):
        """
        out_dir sets the output directory for the simulation results.

        out_dir is a *string* type parameter, specifying the name of the output
        directory in which the model outputs should be saved.
        """
        return self._out_dir

    @out_dir.setter
    def out_dir(self, out_dir):
        self._out_dir = out_dir

    @property
    def verbose(self):
        """
        verbose controls the degree of information printed to the log.

        verbose is an *integer* type parameter, which controls the verbosity
        of the information saved in the log file. A value of 0, the default,
        is the least informative. A value of 1 saves and prints a bit of
        information, and a value of 2 increases the verbosity further.
        """
        return self._verbose

    @verbose.setter
    def verbose(self, verbose):
        self._verbose = verbose

    @property
    def seed(self):
        """
        seed defines the random number seed used for the simulation.

        seed is an *integer* type parameter specifying the random seed value to
        be used for this model run. If unspecified, a random seed is generated
        and used.
        """
        return self._seed

    @seed.setter
    def seed(self, seed):
        self._seed = seed

    @property
    def Length(self):
        """
        Length sets the total length of the domain.

        Length is either an *integer* or a *float*.
        This is the length of the domain (dimension parallel to the inlet
        channel), in **meters**.
        """
        return self._Length

    @Length.setter
    def Length(self, Length):
        if Length <= 0:
            raise ValueError('Length must be a positive number.')
        self._Length = Length

    @property
    def Width(self):
        """
        Width sets the total width of the domain.

        Width is either an *integer* or a *float*.
        This is the width of the domain (dimension perpendicular to the inlet
        channel), in **meters**.
        """
        return self._Width

    @Width.setter
    def Width(self, Width):
        if Width <= 0:
            raise ValueError('Width must be a positive number.')
        self._Width = Width

    @property
    def dx(self):
        """
        dx is the length of the individual cell faces.

        dx is either an *integer* or a *float*.
        This parameter specifies the length of the cell faces in the grid in
        **meters**.
        """
        return self._dx

    @dx.setter
    def dx(self, dx):
        if (dx <= 0) or (dx > self._Length) or (dx > self._Width):
            raise ValueError('dx must be positive and smaller than the' +
                             ' Length and Width parameters.')
        self._dx = dx

    @property
    def L0_meters(self):
        """
        L0_meters defines the thickness of the land along the coast.

        L0 meters is either an *integer* or a *float*.
        Thickness of the land adjacent to the inlet in **meters**.
        This can also be thought of as the length of the inlet channel.
        """
        return self._L0_meters

    @L0_meters.setter
    def L0_meters(self, L0_meters):
        if L0_meters < 0:
            raise ValueError('L0_meters must be a greater than or equal to 0.')
        self._L0_meters = L0_meters

    @property
    def S0(self):
        """
        S0 is the characteristic slope for the delta.

        S0 is either an *integer* or a *float*.
        This sets the characteristic slope for the delta topset.
        This parameter is dimensionless.
        """
        return self._S0

    @S0.setter
    def S0(self, S0):
        self._S0 = S0

    @property
    def itermax(self):
        """
        itermax sets the number of flow routing/free surface iterations.
        """
        return self._itermax

    @itermax.setter
    def itermax(self, itermax):
        if itermax < 0:
            raise ValueError('itermax must be greater than or equal to 0.')
        self._itermax = itermax

    @property
    def Np_water(self):
        """
        Np_water is the number of water parcels simulated.

        Np_water represents the number of "parcels" to split the input water
        discharge into for the reduced-complexity flow routing.
        This parameter must be an *integer*
        """
        return self._Np_water

    @Np_water.setter
    def Np_water(self, Np_water):
        if Np_water <= 0:
            raise ValueError('Np_water must be a positive number.')
        self._Np_water = Np_water

    @property
    def u0(self):
        """
        u0 is a reference velocity value.

        u0 is the characteristic or reference velocity value in units of m/s.
        u0 influences the values of other model parameters such as the maximum
        flow velocity, gamma, and the velocities at which sediments deposit and
        erode. This parameter must be an *integer* or a *float*.
        """
        return self._u0

    @u0.setter
    def u0(self, u0):
        self._u0 = u0

    @property
    def N0_meters(self):
        """
        N0_meters defines the width of the inlet channel in meters.

        N0_meters defines the width of the inlet channel in meters. Therefore,
        this parameter must be a positive *integer* or *float*.
        """
        return self._N0_meters

    @N0_meters.setter
    def N0_meters(self, N0_meters):
        if N0_meters <= 0:
            raise ValueError('N0_meters must be a positive number.')
        self._N0_meters = N0_meters

    @property
    def h0(self):
        """
        h0 is the reference or characteristic water depth in meters.

        h0 is the reference or characteristic water depth in meters. This
        parameter must be an *integer* or *float*. h0 defines
        the depth of water in the inlet channel. One-tenth of the value of h0
        defines the "dry-depth" or the depth at which cells are considered to
        be non-wet (dry).
        """
        return self._h0

    @h0.setter
    def h0(self, h0):
        self._h0 = h0

    @property
    def hb(self):
        """
        hb is the basin depth in meters.

        hb is the basin depth in meters. This parameter must be an *integer*
        or *float*, or *None*. If no value is supplied for this parameter,
        then the value of h0 is used to determine the basin depth (i.e., they
        are the same).
        """
        return self._hb

    @hb.setter
    def hb(self, hb):
        self._hb = hb

    @property
    def H_SL(self):
        """
        H_SL sets the sea level elevation in meters.
        """
        return self._H_SL

    @H_SL.setter
    def H_SL(self, H_SL):
        self._H_SL = H_SL

    @property
    def SLR(self):
        """
        SLR is the sea level rise rate.

        SLR is a parameter for defining the sea level rise rate. SLR is
        specified in units of m/s. When prescribing a SLR rate, it is important
        to remember that pyDeltaRCM simulates bankfull discharge conditions.
        Depending on the flood intermittancy intervals you assume, the
        conversion from "model time" into "real time simulated" may vary.
        """
        return self._SLR

    @SLR.setter
    def SLR(self, SLR):
        self._SLR = SLR

    @property
    def If(self):
        """
        Intermittency factor converting model time to real-world time.

        .. important::

            This parameter has no effect on model operations. It is used in
            determining model run duration.
        """
        return self._If

    @If.setter
    def If(self, If):
        if (If <= 0) or (If > 1):
            raise ValueError('If must be in interval (0, 1].')
        self._If = If

    @property
    def Np_sed(self):
        """
        Np_sed is the number of sediment parcels simulated.

        Np_sed represents the number of "parcels" to split the input sediment
        discharge into for the reduced-complexity sediment routing.
        This parameter must be an *integer*
        """
        return self._Np_sed

    @Np_sed.setter
    def Np_sed(self, Np_sed):
        if Np_sed <= 0:
            raise ValueError('Np_sed must be a positive number.')
        self._Np_sed = Np_sed

    @property
    def f_bedload(self):
        """
        f_bedload is the bedload fraction of the input sediment.

        f_bedload is the fraction of input sediment that is bedload material.
        In pyDeltaRCM, bedload material is coarse grained "sand", and
        suspended load material is fine grained "mud". This parameter must be
        a value between 0 and 1, inclusive.
        """
        return self._f_bedload

    @f_bedload.setter
    def f_bedload(self, f_bedload):
        if (f_bedload < 0) or (f_bedload > 1):
            raise ValueError('Value for f_bedload must be between 0 and 1,'
                             ' inclusive.')
        self._f_bedload = f_bedload

    @property
    def active_layer_thickness(self):
        """
        Active layer thickness for sand fraction.

        The active layer thickness is the depth to which the sand fraction
        values are maintained in the event of erosion from one timestep to the
        next. When erosion exceeds the depth of the active layer, the boundary
        condition is to fill that cell with a value of 0 (maybe this should be
        -1 so that it is clear which cells are not truly mud but unknown
        sediment content based on the active layer thickness?)
        """
        return self._active_layer_thickness

    @active_layer_thickness.setter
    def active_layer_thickness(self, active_layer_thickness):
        if active_layer_thickness is None:
            active_layer_thickness = self._h0 / 2
        elif active_layer_thickness < 0:
            raise ValueError('active_layer thickness must be greater than'
                             ' or equal to 0, cannot be negative.')
        self._active_layer_thickness = active_layer_thickness

    @property
    def C0_percent(self):
        """
        C0_percent is the sediment concentration in the input water supply.

        C0_percent is the prescribed sediment concentration in the input water
        as a percentage (must be equal to or greater than 0).
        """
        return self._C0_percent

    @C0_percent.setter
    def C0_percent(self, C0_percent):
        if C0_percent < 0:
            raise ValueError('C0_percent must be greater than 0.')
        self._C0_percent = C0_percent

    @property
    def Csmooth(self):
        """
        Csmooth is a free surface smoothing parameter.
        """
        return self._Csmooth

    @Csmooth.setter
    def Csmooth(self, Csmooth):
        if Csmooth < 0:
            raise ValueError('Csmooth must be greater than or equal to 0.')
        self._Csmooth = Csmooth

    @property
    def toggle_subsidence(self):
        """
        toggle_subsidence controls whether subsidence is turned on or off.

        If toggle_subsidence is set to `True` then subsidence is turned on.
        Otherwise if toggle_subsidence is set to `False` (the default) then no
        subsidence will occur.
        """
        return self._toggle_subsidence

    @toggle_subsidence.setter
    def toggle_subsidence(self, toggle_subsidence):
        self._toggle_subsidence = toggle_subsidence

    @property
    def subsidence_rate(self):
        """
        subsidence_rate defines the maximum rate of subsidence.
        """
        return self._subsidence_rate

    @subsidence_rate.setter
    def subsidence_rate(self, subsidence_rate):
        self._subsidence_rate = subsidence_rate

    @property
    def start_subsidence(self):
        """
        start_subsidence defines the start time at which subsidence begins.
        """
        return self._start_subsidence

    @start_subsidence.setter
    def start_subsidence(self, start_subsidence):
        self._start_subsidence = start_subsidence

    @property
    def save_eta_figs(self):
        """
        save_eta_figs controls whether or not figures of topography are saved.
        """
        return self._save_eta_figs

    @save_eta_figs.setter
    def save_eta_figs(self, save_eta_figs):
        if (save_eta_figs is True) and \
          ('eta' not in self._save_fig_list.keys()):
            self._save_fig_list['eta'] = ['eta']
        elif ((save_eta_figs is False) and
              ('eta' in self._save_fig_list.keys())):
            del self._save_fig_list['eta']
        self._save_eta_figs = save_eta_figs

    @property
    def save_stage_figs(self):
        """
        save_stage_figs controls whether or not stage figures are saved.
        """
        return self._save_stage_figs

    @save_stage_figs.setter
    def save_stage_figs(self, save_stage_figs):
        if (save_stage_figs is True) and \
          ('stage' not in self._save_fig_list.keys()):
            self._save_fig_list['stage'] = ['stage']
        elif ((save_stage_figs is False) and
              ('stage' in self._save_fig_list.keys())):
            del self._save_fig_list['stage']
        self._save_stage_figs = save_stage_figs

    @property
    def save_depth_figs(self):
        """
        save_depth_figs controls saving of water depth figures.
        """
        return self._save_depth_figs

    @save_depth_figs.setter
    def save_depth_figs(self, save_depth_figs):
        if (save_depth_figs is True) and \
          ('depth' not in self._save_fig_list.keys()):
            self._save_fig_list['depth'] = ['depth']
        elif ((save_depth_figs is False) and
              ('depth' in self._save_fig_list.keys())):
            del self._save_fig_list['depth']
        self._save_depth_figs = save_depth_figs

    @property
    def save_discharge_figs(self):
        """
        save_discharge_figs controls saving of water discharge figures.
        """
        return self._save_discharge_figs

    @save_discharge_figs.setter
    def save_discharge_figs(self, save_discharge_figs):
        if (save_discharge_figs is True) and \
          ('discharge' not in self._save_fig_list.keys()):
            self._save_fig_list['discharge'] = ['qw']
        elif ((save_discharge_figs is False) and
              ('discharge' in self._save_fig_list)):
            del self._save_fig_list['discharge']
        self._save_discharge_figs = save_discharge_figs

    @property
    def save_velocity_figs(self):
        """
        save_velocity_figs controls saving of water velocity figures.
        """
        return self._save_velocity_figs

    @save_velocity_figs.setter
    def save_velocity_figs(self, save_velocity_figs):
        if (save_velocity_figs is True) and \
          ('velocity' not in self._save_fig_list.keys()):
            self._save_fig_list['velocity'] = ['uw']
        elif ((save_velocity_figs is False) and
              ('velocity' in self._save_fig_list.keys())):
            del self._save_fig_list['sedflux']
        self._save_velocity_figs = save_velocity_figs

    @property
    def save_sedflux_figs(self):
        """
        save_sedflux_figs controls saving of sediment flux figures.
        """
        return self._save_sedflux_figs

    @save_sedflux_figs.setter
    def save_sedflux_figs(self, save_sedflux_figs):
        if (save_sedflux_figs is True) and \
          ('sedflux' not in self._save_fig_list.keys()):
            self._save_fig_list['sedflux'] = ['qs']
        elif ((save_sedflux_figs is False) and
              ('sedflux' in self._save_fig_list)):
            del self._save_fig_list['sedflux']
        self._save_sedflux_figs = save_sedflux_figs

    @property
    def save_sandfrac_figs(self):
        """
        save_sandfrac_figs controls whether or not figures of the sediment
        surface (i.e., the bed) sand fraction are saved.
        """
        return self._save_sandfrac_figs

    @save_sandfrac_figs.setter
    def save_sandfrac_figs(self, save_sandfrac_figs):
        if (save_sandfrac_figs is True) and \
          ('sandfrac' not in self._save_fig_list.keys()):
            self._save_fig_list['sandfrac'] = ['sand_frac']
        elif ((save_sandfrac_figs is False) and
              ('sandfrac' in self._save_fig_list)):
            del self._save_fig_list['sandfrac']
        self._save_sandfrac_figs = save_sandfrac_figs

    @property
    def save_figs_sequential(self):
        """
        save_figs_sequential sets how figures are to be saved.

        save_figs_sequential is a *boolean* parameter that can be set to True
        or False. If True, then for each figure saving parameter set to True
        (e.g. :attr:`save_velocity_figs`) a new figure will be saved at an
        inteval of :attr:`save_dt`. The file names of the figures will
        correspond to the model timestep at which they were saved. If instead,
        the save_figs_sequential parameter is set to False, then only the
        latest figure will be kept, and at each *save_dt* time, the figure file
        will be overwritten using the current model status.
        """
        return self._save_figs_sequential

    @save_figs_sequential.setter
    def save_figs_sequential(self, save_figs_sequential):
        self._save_figs_sequential = save_figs_sequential

    @property
    def save_metadata(self):
        """
        save_metadata explicit control on whether or not metadata is saved.

        save_metadata is a boolean that can be manually togged on (True) to
        ensure metadata is saved to disk even if no other output information is
        being saved. If any grids or strata are being saved, then metadata
        saving will be turned on automatically, even if this parameter is set
        to False. Metadata associated with pyDeltaRCM are 1-D arrays
        (vectors) and 0-D arrays (floats and integers) primarily containing
        information about the domain and the inlet conditions for a given model
        run.
        """
        return self._save_metadata

    @save_metadata.setter
    def save_metadata(self, save_metadata):
        self._save_metadata = save_metadata

    @property
    def save_eta_grids(self):
        """
        save_eta_grids controls whether or not topography information is saved.
        """
        return self._save_eta_grids

    @save_eta_grids.setter
    def save_eta_grids(self, save_eta_grids):
        if (save_eta_grids is True) and \
          ('eta' not in self._save_var_list.keys()):
            self._save_var_list['eta'] = ['eta', 'meters', 'f4',
                                          self._netcdf_coords]
        elif ((save_eta_grids is False) and
              ('eta' in self._save_var_list.keys())):
            del self._save_var_list['eta']
        self._save_eta_grids = save_eta_grids

    @property
    def save_stage_grids(self):
        """
        save_stage_grids controls whether or not stage information is saved.
        """
        return self._save_stage_grids

    @save_stage_grids.setter
    def save_stage_grids(self, save_stage_grids):
        if (save_stage_grids is True) and \
          ('stage' not in self._save_var_list.keys()):
            self._save_var_list['stage'] = ['stage', 'meters', 'f4',
                                            self._netcdf_coords]
        elif ((save_stage_grids is False) and
              ('stage' in self._save_var_list.keys())):
            del self._save_var_list['stage']
        self._save_stage_grids = save_stage_grids

    @property
    def save_depth_grids(self):
        """
        save_depth_grids controls whether or not depth information is saved.
        """
        return self._save_depth_grids

    @save_depth_grids.setter
    def save_depth_grids(self, save_depth_grids):
        if (save_depth_grids is True) and \
          ('depth' not in self._save_var_list.keys()):
            self._save_var_list['depth'] = ['depth', 'meters', 'f4',
                                            self._netcdf_coords]
        elif ((save_depth_grids is False) and
              ('depth' in self._save_var_list.keys())):
            del self._save_var_list['depth']
        self._save_depth_grids = save_depth_grids

    @property
    def save_discharge_grids(self):
        """
        save_discharge_grids controls saving of water discharge information.
        """
        return self._save_discharge_grids

    @save_discharge_grids.setter
    def save_discharge_grids(self, save_discharge_grids):
        if (save_discharge_grids is True) and \
          ('discharge' not in self._save_var_list.keys()):
            self._save_var_list['discharge'] = ['qw',
                                                'cubic meters per second',
                                                'f4',
                                                self._netcdf_coords]
        elif ((save_discharge_grids is False) and
              ('discharge' in self._save_var_list.keys())):
            del self._save_var_list['discharge']
        self._save_discharge_grids = save_discharge_grids

    @property
    def save_velocity_grids(self):
        """
        save_velocity_grids controls saving of water velocity information.
        """
        return self._save_velocity_grids

    @save_velocity_grids.setter
    def save_velocity_grids(self, save_velocity_grids):
        if (save_velocity_grids is True) and \
          ('velocity' not in self._save_var_list.keys()):
            self._save_var_list['velocity'] = ['uw', 'meters per second', 'f4',
                                               self._netcdf_coords]
        elif ((save_velocity_grids is False) and
              ('velocity' in self._save_var_list.keys())):
            del self._save_var_list['velocity']
        self._save_velocity_grids = save_velocity_grids

    @property
    def save_sedflux_grids(self):
        """
        save_sedflux_grids controls saving of sediment discharge information.
        """
        return self._save_sedflux_grids

    @save_sedflux_grids.setter
    def save_sedflux_grids(self, save_sedflux_grids):
        if (save_sedflux_grids is True) and \
          ('sedflux' not in self._save_var_list.keys()):
            self._save_var_list['sedflux'] = ['qs', 'cubic meters per second',
                                              'f4',
                                              self._netcdf_coords]
        elif ((save_sedflux_grids is False) and
              ('sedflux' in self._save_var_list.keys())):
            del self._save_var_list['sedflux']
        self._save_sedflux_grids = save_sedflux_grids

    @property
    def save_sandfrac_grids(self):
        """
        save_sandfrac_grids controls whether or not the sediment
        surface (i.e., the bed) sand fraction is saved.
        """
        return self._save_sandfrac_grids

    @save_sandfrac_grids.setter
    def save_sandfrac_grids(self, save_sandfrac_grids):
        if (save_sandfrac_grids is True) and \
          ('sandfrac' not in self._save_var_list.keys()):
            self._save_var_list['sandfrac'] = ['sand_frac', 'fraction', 'f4',
                                               self._netcdf_coords]
        elif ((save_sandfrac_grids is False) and
              ('sandfrac' in self._save_var_list.keys())):
            del self._save_var_list['sandfrac']
        self._save_sandfrac_grids = save_sandfrac_grids

    @property
    def save_discharge_components(self):
        """
        save_discharge_components controls saving of x-y discharge components.
        """
        return self._save_discharge_components

    @save_discharge_components.setter
    def save_discharge_components(self, save_discharge_components):
        if (save_discharge_components is True):
            if ('discharge_x' not in self._save_var_list.keys()):
                self._save_var_list['discharge_x'] = [
                    'qx', 'cubic meters per second', 'f4',
                    self._netcdf_coords]
            if ('discharge_y' not in self._save_var_list.keys()):
                self._save_var_list['discharge_y'] = [
                    'qy', 'cubic meters per second', 'f4',
                    self._netcdf_coords]
        elif (save_discharge_components is False):
            if ('discharge_x' in self._save_var_list.keys()):
                del self._save_var_list['discharge_x']
            if ('discharge_y' in self._save_var_list.keys()):
                del self._save_var_list['discharge_y']
        self._save_discharge_components = save_discharge_components

    @property
    def save_velocity_components(self):
        """
        save_velocity_components controls saving of x-y velocity components.
        """
        return self._save_velocity_components

    @save_velocity_components.setter
    def save_velocity_components(self, save_velocity_components):
        if (save_velocity_components is True):
            if ('velocity_x' not in self._save_var_list.keys()):
                self._save_var_list['velocity_x'] = [
                    'ux', 'meters per second', 'f4',
                    self._netcdf_coords]
            if ('velocity_y' not in self._save_var_list.keys()):
                self._save_var_list['velocity_y'] = [
                    'uy', 'meters per second', 'f4',
                    self._netcdf_coords]
        elif (save_velocity_components is False):
            if ('velocity_x' in self._save_var_list.keys()):
                del self._save_var_list['velocity_x']
            if ('velocity_y' in self._save_var_list.keys()):
                del self._save_var_list['velocity_y']
        self._save_velocity_components = save_velocity_components

    @property
    def save_dt(self):
        """
        save_dt defines the saving interval in seconds.
        """
        return self._save_dt

    @save_dt.setter
    def save_dt(self, save_dt):
        self._save_dt = save_dt

    @property
    def checkpoint_dt(self):
        """
        checkpoint_dt defines the interval to create checkpoint information.
        """
        return self._checkpoint_dt

    @checkpoint_dt.setter
    def checkpoint_dt(self, checkpoint_dt):
        if checkpoint_dt is None:
            checkpoint_dt = self._save_dt
        self._checkpoint_dt = checkpoint_dt

    @property
    def save_checkpoint(self):
        """
        save_checkpoint controls saving of model checkpoint information.
        """
        return self._save_checkpoint

    @save_checkpoint.setter
    def save_checkpoint(self, save_checkpoint):
        self._save_checkpoint = save_checkpoint

    @property
    def resume_checkpoint(self):
        """
        resume_checkpoint controls loading of a checkpoint if run is resuming.

        When setting this option in the YAML or command line, you can specify
        a `bool` (e.g., `resume_checkpoint: True`) which will search in the
        :obj:`out_dir` directory for a file named ``checkpoint.npz``.
        Alternatively, you can specify an alternative folder to search for the
        checkpoint *as a string* (e.g., `resume_checkpoint:
        '/some/other/path'`).
        """
        return self._resume_checkpoint

    @resume_checkpoint.setter
    def resume_checkpoint(self, resume_checkpoint):
        if isinstance(resume_checkpoint, str):
            self._checkpoint_folder = resume_checkpoint
            self._resume_checkpoint = True
        else:
            self._checkpoint_folder = self.prefix
            self._resume_checkpoint = resume_checkpoint

    @property
    def omega_sfc(self):
        """
        omega_sfc is a water surface underrelaxation parameter.
        """
        return self._omega_sfc

    @omega_sfc.setter
    def omega_sfc(self, omega_sfc):
        self._omega_sfc = omega_sfc

    @property
    def omega_flow(self):
        """
        omega_flow is a flow velocity underrelaxation parameter.
        """
        return self._omega_flow

    @omega_flow.setter
    def omega_flow(self, omega_flow):
        self._omega_flow = omega_flow

    @property
    def Nsmooth(self):
        """
        Nsmooth defines the number of times the water surface is smoothed.
        """
        return self._Nsmooth

    @Nsmooth.setter
    def Nsmooth(self, Nsmooth):
        self._Nsmooth = Nsmooth

    @property
    def theta_water(self):
        """
        theta_water is the exponent of depth dependence for weighted routing.

        For the routing of the water parcels, the dependence of the random walk
        on local water depth can be modulated by varying this parameter value.
        As theta_water gets larger, the importance of the water depth to the
        weighting scheme grows.

        .. note::
           The value of *theta_water* also influences the values of
           :attr:`coeff_theta_sand` and :attr:`coeff_theta_mud` which are
           coefficients that are multiplied by *theta_water* to set the theta
           values for the sand and mud routing respectively.
        """
        return self._theta_water

    @theta_water.setter
    def theta_water(self, theta_water):
        self._theta_water = theta_water

    @property
    def coeff_theta_sand(self):
        """
        coeff_theta_sand is the coefficient applied to theta for sand routing.

        coeff_theta_sand is a coefficient applied to the :attr:`theta_water`
        attribute to define the value of theta for sand routing. Theta is
        the exponent applied to the local water depth and is used to weight
        the random walk. For mud and sand these weighting rules vary, which is
        why the :attr:`theta_water` term is multiplied by these coefficients:
        :attr:`coeff_theta_mud` and :attr:`coeff_theta_sand`.
        """
        return self._coeff_theta_sand

    @coeff_theta_sand.setter
    def coeff_theta_sand(self, coeff_theta_sand):
        self._coeff_theta_sand = coeff_theta_sand

    @property
    def coeff_theta_mud(self):
        """
        coeff_theta_mud is the coefficient applied to theta for mud routing.

        coeff_theta_mud is a coefficient applied to the :attr:`theta_water`
        attribute to define the value of theta for mud routing. Theta is
        the exponent applied to the local water depth and is used to weight
        the random walk. For mud and sand these weighting rules vary, which is
        why the :attr:`theta_water` term is multiplied by these coefficients:
        :attr:`coeff_theta_mud` and :attr:`coeff_theta_sand`.
        """
        return self._coeff_theta_mud

    @coeff_theta_mud.setter
    def coeff_theta_mud(self, coeff_theta_mud):
        self._coeff_theta_mud = coeff_theta_mud

    @property
    def beta(self):
        """
        beta is the bedload transport capacity exponent.

        beta is an exponent on the bedload transport terms that is applied to
        the local velocity and threshold velocity terms.
        """
        return self._beta

    @beta.setter
    def beta(self, beta):
        self._beta = beta

    @property
    def sed_lag(self):
        """
        sed_lag is a "sedimentation lag" parameter.

        sed_lag influences the properties of mud deposition by controlling how
        much mud is deposited when the flow velocity is below the threshold
        for mud deposition (:attr:`coeff_U_dep_mud`).
        """
        return self._sed_lag

    @sed_lag.setter
    def sed_lag(self, sed_lag):
        self._sed_lag = sed_lag

    @property
    def coeff_U_dep_mud(self):
        """
        coeff_U_dep_mud is the threshold velocity for mud sediment deposition.

        coeff_U_dep_mud sets the threshold velocity for mud to be depoisted.
        The smaller this parameter is, the longer a mud parcel can travel
        before losing all of its mud volume.
        """
        return self._coeff_U_dep_mud

    @coeff_U_dep_mud.setter
    def coeff_U_dep_mud(self, coeff_U_dep_mud):
        self._coeff_U_dep_mud = coeff_U_dep_mud

    @property
    def coeff_U_ero_mud(self):
        """
        coeff_U_ero_mud is the mud erosion velocity threshold coefficient.

        coeff_U_ero_mud sets the threshold velocity for erosion of the mud
        (suspended load) sediment. The higher this value is, the more difficult
        it is to erode mud deposits.
        """
        return self._coeff_U_ero_mud

    @coeff_U_ero_mud.setter
    def coeff_U_ero_mud(self, coeff_U_ero_mud):
        self._coeff_U_ero_mud = coeff_U_ero_mud

    @property
    def coeff_U_ero_sand(self):
        """
        coeff_U_ero_sand is the sand erosion velocity threshold coefficient.

        coeff_U_ero_sand sets the threshold velocity for sediment erosion of
        the sand (bedload) sediment. The higher this value is, the more
        difficult it is to erode the bed.
        """
        return self._coeff_U_ero_sand

    @coeff_U_ero_sand.setter
    def coeff_U_ero_sand(self, coeff_U_ero_sand):
        self._coeff_U_ero_sand = coeff_U_ero_sand

    @property
    def alpha(self):
        """
        alpha is the topographic diffusion coefficient.

        alpha is the coefficient used for topographic diffusion. It controls
        both the cross-slope sediment flux as well as bank erodability.
        """
        return self._alpha

    @alpha.setter
    def alpha(self, alpha):
        self._alpha = alpha

    @property
    def stepmax(self):
        """
        `stepmax` is the maximum number of jumps a parcel can make.

        The maximum number of jumps a parcel (water or sediment) can take is
        determined by `stepmax`. If the parcel reaches the oceanic boundary
        before `stepmax` is reached, then this condition is not invoked.
        However if the parcel does not reach the oceanic bounary but does
        reach `stepmax` jumps, then the parcel will just stop moving and
        disappear.

        If `stepmax` is not specified in the yaml, the default value assigned
        is 2 times the perimeter of the domain (2 * (self.L + self.W)).

        .. note::
           If `stepmax` is set too low and many parcels reach the `stepmax`
           condition during routing, then there may be a considerable
           amount of sediment 'missing' from the system as any sediment in
           a parcel that hits the `stepmax` threshold disappears from the
           simulation.
        """
        return self._stepmax

    @stepmax.setter
    def stepmax(self, stepmax):
        self._stepmax = stepmax

    @property
    def clobber_netcdf(self):
        """
        Allows overwriting (clobbering) of an existing netCDF output file.

        Default behavior, clobber_netcdf: False, is for the model to raise an
        error during initialization if a netCDF output file is found in the
        location specified by :attr:`out_dir`. If the clobber_netcdf parameter
        is instead set to True, then if there is an existing netCDF output file
        in the target folder, it will be "clobbered" or overwritten.
        """
        return self._clobber_netcdf

    @clobber_netcdf.setter
    def clobber_netcdf(self, clobber_netcdf):
        self._clobber_netcdf = clobber_netcdf

    @property
    def legacy_netcdf(self):
        """Enable output in legacy netCDF format.

        .. note:: new in `v2.1.0`.

        Default behavior, legacy_netcdf: False, is for the model to use the
        new `v2.1.0` output netCDF format. The updated format is configured
        to match the input expected by `xarray`, which eases interaction with
        model outputs. The change in format is from inconsistently named
        dimensions and *coordinate variables*, to homogeneous definitions.
        Also, the legacy format specified the variables `x` and `y` as 2d
        grids, whereas the updated format uses 1d coordinate arrays.
        
        .. important::

            There are no changes to the dimensionality of data such as bed
            elevation or velocity, only the metadata specifying the location of
            the data are changed.

        +-------------+-------------------+---------------------------------+
        |             | default           | legacy                          |
        +=============+===================+=================================+
        | dimensions  | `time`, `x`, `y`  | `total_time`, `length`, `width` |
        +-------------+-------------------+---------------------------------+
        | variables   | `time`, `x`, `y`  | `time`, `y`, `x`; x, y as 2D    |
        +-------------+-------------------+---------------------------------+
        | data        | `t-x-y` array     | `t-y-x` array                   |
        +-------------+-------------------+---------------------------------+

        .. hint::

            If you are beginning a new project, use `legacy_netcdf == False`,
            and update scripts accordingly.
        """
        return self._legacy_netcdf

    @legacy_netcdf.setter
    def legacy_netcdf(self, legacy_netcdf):
        self._legacy_netcdf = legacy_netcdf

    @property
    def time(self):
        """Elapsed model time in seconds.
        """
        return self._time

    @property
    def dt(self):
        """The time step.

        The value of the timestep (:math:`\\Delta t`) is a balance between
        computation efficiency and model stability [1]_. The
        :ref:`reference-volume`, which characterizes the
        volume of an inlet-channel cell, and the sediment discharge to the
        model domain scale the timestep as:

        .. math::

            \\Delta t = \\dfrac{0.1 {N_0}^2 V_0}{Q_{s0}}

        where :math:`Q_{s0}` is the sediment discharge into the model domain
        (m:sup:`3`/s), :math:`V_0` is the reference volume, and :math:`N_0` is
        the number of cells across the channel (determined by :obj:`N0_meters`
        and :obj:`dx`).

        .. [1] A reduced-complexity model for river delta formation – Part 1:
               Modeling deltas with channel dynamics, M. Liang, V. R. Voller,
               and C. Paola, Earth Surf. Dynam., 3, 67–86, 2015.
               https://doi.org/10.5194/esurf-3-67-2015

        Raises
        ------
        UserWarning
            If a very small timestep is configured.
        """
        return self._dt

    @property
    def time_step(self):
        """Alias for :obj:`dt`.
        """
        return self._dt

    @time_step.setter
    def time_step(self, new_time_step):
        if new_time_step * self.init_Np_sed < 100:
            warnings.warn(UserWarning('Using a very small time step, '
                                      'Delta might evolve very slowly.'))

        if self.toggle_subsidence:
            self.sigma = (self.sigma / self._dt) * new_time_step

        self._dt = new_time_step
        self.init_sediment_routers()

    @property
    def time_iter(self):
        """Number of time iterations.

        The number of times the :obj:`update` method has been called.
        """
        return self._time_iter

    @property
    def save_time_since_data(self):
        """Time since data last output.

        The elapsed time (seconds) since data was output with
        :obj:`output_data`.
        """
        return self._save_time_since_data

    @property
    def save_time_since_checkpoint(self):
        """Time since last data checkpoint was saved."""
        return self._save_time_since_checkpoint

    @property
    def save_iter(self):
        """Number of times data has been saved."""
        return self._save_iter

    @property
    def channel_flow_velocity(self):
        """Get channel flow velocity."""
        return self.u0

    @channel_flow_velocity.setter
    def channel_flow_velocity(self, new_u0):
        self.u0 = new_u0
        self.create_boundary_conditions()
        self.init_sediment_routers()

    @property
    def channel_width(self):
        """Get channel width."""
        return self.N0 * self._dx

    @channel_width.setter
    def channel_width(self, new_N0_meters):
        self.N0_meters = new_N0_meters
        self.create_boundary_conditions()
        self.init_sediment_routers()
        if self.channel_width != new_N0_meters:
            warnings.warn(UserWarning(
                'Channel width was updated to {0} m, rather than input {1},'
                'due to grid resolution or imposed domain '
                'restrictions.'.format(self.channel_width, new_N0_meters)))

    @property
    def channel_flow_depth(self):
        """Get channel flow depth."""
        return self.h0

    @channel_flow_depth.setter
    def channel_flow_depth(self, new_d):
        self.h0 = new_d
        self.create_boundary_conditions()
        self.init_sediment_routers()

    @property
    def sea_surface_mean_elevation(self):
        """Get sea surface mean elevation."""
        return self.H_SL

    @sea_surface_mean_elevation.setter
    def sea_surface_mean_elevation(self, new_se):
        self.H_SL = new_se

    @property
    def sea_surface_elevation_change(self):
        """Get rate of change of sea surface elevation, per timestep."""
        return self.SLR

    @sea_surface_elevation_change.setter
    def sea_surface_elevation_change(self, new_SLR):
        self.SLR = new_SLR

    @property
    def bedload_fraction(self):
        """Get bedload fraction."""
        return self.f_bedload

    @bedload_fraction.setter
    def bedload_fraction(self, new_u0):
        self.f_bedload = new_u0

    @property
    def influx_sediment_concentration(self):
        """Get influx sediment concentration."""
        return self.C0_percent

    @influx_sediment_concentration.setter
    def influx_sediment_concentration(self, new_C0):
        self.C0_percent = new_C0 * 100
        self.create_boundary_conditions()
        self.init_sediment_routers()

    @property
    def sea_surface_elevation(self):
        """Get stage."""
        return self.stage

    @property
    def water_depth(self):
        """Get depth."""
        return self.depth

    @property
    def bed_elevation(self):
        """Get bed elevation."""
        return self.eta

    @property
    def sand_frac_bc(self):
        """Sand fraction boundary condition."""
        return self._sand_frac_bc

    @sand_frac_bc.setter
    def sand_frac_bc(self, sand_frac_bc):
        self._sand_frac_bc = sand_frac_bc



import matplotlib.pyplot as plt
import numpy as np

import _base

fig, ax = plt.subplots()
plt.subplots_adjust(left=0.2)
ax.plot(_base.time, _base.hydrograph)
ax.axhline(y=2.5, color='k', ls='--')
ax.fill_between(np.arange(365), 2.5 * np.ones((len(_base.time),)),
                _base.hydrograph, where=_base.flood, alpha=0.4)
ax.set_yticks([1.05, 2.5])
ax.set_yticklabels(['base flow', 'bankfull flow'])
ax.set_xlabel('calendar day')
ax.set_ylim(0.9, 3.1)

bbox = dict(boxstyle="round", fc="0.8")
_cs0 = "angle,angleA=-10,angleB=-60,rad=10"
ax.annotate('start event', xy=(95, 1.1), xytext=(0, 1.4),
            bbox=bbox, arrowprops=dict(arrowstyle="->", connectionstyle=_cs0))
_cs1 = "angle,angleA=10,angleB=60,rad=10"
ax.annotate('end event', xy=(150, 1.1), xytext=(170, 1.4),
            bbox=bbox, arrowprops=dict(arrowstyle="->", connectionstyle=_cs1))
_cs2 = "angle,angleA=-10,angleB=60,rad=10"
ax.annotate('flooding duration', xy=(110, 2.7), xytext=(150, 2.7),
            bbox=bbox, arrowprops=dict(arrowstyle="->", connectionstyle=_cs2))

plt.show()



import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

import pyDeltaRCM


# filter out the warning raised about no netcdf being found
warnings.filterwarnings("ignore", category=UserWarning)


n = 1
cm = matplotlib.cm.get_cmap('tab10')


# init delta model
with pyDeltaRCM.shared_tools._docs_temp_directory() as output_dir:
    delta = pyDeltaRCM.DeltaModel(
        out_dir=output_dir,
        resume_checkpoint='../../_resources/checkpoint')
_shp = delta.eta.shape


delta.init_water_iteration()
delta.run_water_iteration()
delta.compute_free_surface()


# pidx = 60

# sfc_visit, sfc_sum = pyDeltaRCM.water_tools._accumulate_free_surface_walks(
#     delta.free_surf_walk_inds, delta.free_surf_flag, delta.cell_type,
#     delta.uw, delta.ux, delta.uy, delta.depth,
#     delta._dx, delta._u0, delta.h0, delta._H_SL, delta._S0)


Hnew = np.full_like(delta.sfc_visit, np.nan)
Hnew[delta.sfc_visit > 0] = (delta.sfc_sum[delta.sfc_visit > 0] /
                             delta.sfc_visit[delta.sfc_visit > 0])


# stage_new = delta.eta + delta.depth
# stage_new[delta.sfc_visit > 0] = (delta.sfc_sum[delta.sfc_visit > 0] /
#                             delta.sfc_visit[delta.sfc_visit > 0])


fig, ax = plt.subplots(2, 2, sharex=True, sharey=True, figsize=(9, 4))

pyDeltaRCM.debug_tools.plot_domain(
    delta.sfc_visit, ax=ax[0, 0], grid=False, cmap='Greys',
    label='sfc_visit (-)')
pyDeltaRCM.debug_tools.plot_domain(
    delta.sfc_sum, ax=ax[0, 1], grid=False, cmap='Blues',
    label='sfc_sum (m)')

pyDeltaRCM.debug_tools.plot_domain(
    Hnew, ax=ax[1, 0], grid=False,
    label='computed Hnew (m)')
pyDeltaRCM.debug_tools.plot_domain(
    delta.Hnew, ax=ax[1, 1], grid=False,
    label='stage (m)')

fig.show()



import warnings

import numpy as np
import matplotlib.pyplot as plt

import pyDeltaRCM

# filter out the warning raised about no netcdf being found
warnings.filterwarnings("ignore", category=UserWarning)


# init delta model
with pyDeltaRCM.shared_tools._docs_temp_directory() as output_dir:
    delta = pyDeltaRCM.DeltaModel(
        out_dir=output_dir,
        resume_checkpoint='../../_resources/checkpoint')

_shp = delta.eta.shape

_eta_before = np.copy(delta.eta)

delta.route_all_mud_parcels()

_eta_after = np.copy(delta.eta)


# set up axis
fig, ax = plt.subplots()

# fill in axis2
_diff = _eta_after - _eta_before
pyDeltaRCM.debug_tools.plot_domain(
    _diff, ax=ax, grid=False, cmap='RdBu',
    vmin=-0.1, vmax=0.1)
ax.set_title('bed elevation change (m) \n due to mud parcels')


plt.tight_layout()
plt.show()



import warnings

import matplotlib
import matplotlib.pyplot as plt

import pyDeltaRCM


# filter out the warning raised about no netcdf being found
warnings.filterwarnings("ignore", category=UserWarning)


n = 1
cm = matplotlib.cm.get_cmap('tab10')


# init delta model
with pyDeltaRCM.shared_tools._docs_temp_directory() as output_dir:
    delta = pyDeltaRCM.DeltaModel(
        out_dir=output_dir,
        resume_checkpoint='../../_resources/checkpoint')
_shp = delta.eta.shape


delta.init_water_iteration()
delta.run_water_iteration()


fig, ax = plt.subplots(1, 2, sharex=True, sharey=True, figsize=(9, 3))

# fill in axis
pyDeltaRCM.debug_tools.plot_domain(
    delta.eta, ax=ax[0], grid=False, cmap='cividis',
    label='bed elevation (m)')
pyDeltaRCM.debug_tools.plot_domain(
    delta.eta, ax=ax[1], grid=False, cmap='cividis',
    label='bed elevation (m)')
delta.show_line(delta.free_surf_walk_inds[::10, :].T, 'k-',
                ax=ax[1], alpha=0.1,
                multiline=True, nozeros=True)

fig.show()



import os
import sys
import argparse
import abc
import time
import platform

import itertools
from pathlib import Path
import warnings

import multiprocessing

import yaml
import numpy as np

from . import shared_tools
from .model import DeltaModel as BaseDeltaModel


_ver = ' '.join(('pyDeltaRCM', shared_tools._get_version()))


class BasePreprocessor(abc.ABC):
    """Base preprocessor class.

    Defines a prelimiary yaml reading, then handles the YAML "meta" tag
    parsing, model instatiation, and job running.

    Subclasses create the high-level command line API
    and the high-level python API.

    .. note::

        You probably do not need to interact with this class direclty. Check
        out the Python API.

    """
    def __init__(self):
        """Initialize the base preprocessor.
        """
        self._input_file = None  # initialize empty

        self._has_ensemble = False
        self._has_set = False
        self._has_matrix = False

        self._file_list = []    # list of yaml files for jobs
        self._config_list = []  # list of dict configs for jobs
        self._job_list = []     # list of _Job objects

        self._dryrun = False

        self._is_completed = False

    @property
    def file_list(self):
        """File list.

        A list of `Path` to input YAML files for jobs constructed by the
        preprocessor.
        """
        return self._file_list

    @property
    def config_list(self):
        """Configuration list.

        A list of `dict` containing the input configurations for jobs
        constructed by the preprocessor.
        """
        return self._config_list

    @property
    def job_list(self):
        """Job list.

        A list of jobs, type :obj:`_SerialJob` or :obj:`_ParallelJob`, from
        the jobs executed by the preprocessor.

        .. note::

            This will be an empty list before :obj:`run_jobs` has been called.
        """
        return self._job_list

    def open_input_file_to_dict(self, input_file):
        """Open an input file and convert it to a dictionary.

        This method is used by subclassing Preprocessors to complete the first
        import of the yaml configuration.

        Parameters
        ----------
        input_file
            Path to the input file as string, or Pathlib `Path`.

        Returns
        -------
        yaml_dict
            Input file converted to a dictionary.
        """
        # handle complex preprocessor cases where already is a dict
        if isinstance(input_file, dict):
            yaml_dict = input_file

        # simple cases will be string/path
        elif (isinstance(input_file, str) or
              isinstance(input_file, Path)):
            yaml_dict = self._open_yaml(input_file)

        # something wrong with input
        else:
            raise ValueError('Invalid input file argument.')

        return yaml_dict

    def construct_file_list(self):
        """Construct the file list.

        The job list is constructed by expanding the various multi-job
        specifications. For example, `matrix`, `set`, and
        ensemble yaml files are created in this stage.
        """
        # complete a preliminary parsing of the yaml for high-level controls
        self._prelim_config_parsing()

        # if there is an ensemble specification, expand it first
        if self._has_ensemble:
            self._expand_ensemble()

        if self._has_ensemble or self._has_set or self._has_matrix:
            self._prepare_multijob_output()

        # if there is a matrix or set specification
        if self._has_matrix:
            self._expand_matrix()  # creates self.file_list
        elif self._has_set:
            self._expand_set()
        # otherwise convert to a simple list on input file
        else:
            self._file_list = [self._input_file]
            self._config_list = [self.config_dict]

        # write the job configs to file, if needed
        if self._has_ensemble or self._has_set or self._has_matrix:
            self._write_job_configs()

    def _prelim_config_parsing(self):
        """Preliminary configuration parsing.

        Extract ``.yml`` file (``self.input_file``) into a dictionary, if
        provided. This dictionary provides a few keys used throughout the
        code.

        Here, we check whether the file is valid yaml,
        and place it into a dictionary.

        Additionally, set the ``self._has_matrix`` flag, which is used in the
        :meth:`_expand_matrix`.

        """
        if 'set' in self.config_dict.keys():
            # set found
            self._has_set = True
            # can't specify matrix with set
            if ('matrix' in self.config_dict.keys()):
                raise ValueError(
                    'Cannot use "matrix" and "set" expansion together.')

        if 'ensemble' in self.config_dict.keys():
            # ensemble found
            self._has_ensemble = True

        if 'matrix' in self.config_dict.keys():
            # matrix found
            self._has_matrix = True

        if 'verbose' in self.config_dict.keys():
            self.verbose = self.config_dict['verbose']
        else:
            self.verbose = 0

        if 'dryrun' in self.config_dict.keys():
            self._dryrun = self.config_dict['dryrun']

    def _open_yaml(self, input_file):
        """Safely open, read, and close a yaml file.

        Parameters
        ----------
        input file
            string or path to file

        Returns
        -------
        yaml_dict
            yaml file, as a Python dict
        """
        if (input_file is None):
            return {}  # return an empty dict
        else:
            # get the special loader from the shared tools
            loader = shared_tools.custom_yaml_loader()

            # open the file with the loader
            user_file = open(input_file, mode='r')
            yaml_dict = yaml.load(user_file, Loader=loader)
            user_file.close()
            return yaml_dict

    def _prepare_multijob_output(self):
        if 'out_dir' not in self.config_dict.keys():
            raise ValueError(
                'You must specify "out_dir" in YAML to use any '
                'multi-job expansion tools.')

        self._jobs_root = self.config_dict['out_dir']

        # create directory at root
        p = Path(self._jobs_root)
        if p.is_dir():
            if 'resume_checkpoint' in self.config_dict and \
              self.config_dict['resume_checkpoint'] is True:
                pass
            else:
                raise FileExistsError(
                    'Job output directory (%s) already exists.' % str(p))
        else:
            p.mkdir()

    def _expand_ensemble(self):
        """Create ensemble random seeds and put into matrix.

        Ensemble expansion is implemented as a special class of matrix/set
        expansion where:

            * if set specification is used, each set is duplicated the
              specified number of times, and a seed is added to each set
            * otherwise, the matrix key is set for the `seed` field of the
              model

        This approach ensures that the runs will have different
        outcomes, while allowing all other parameters to remain fixed, and
        supporting additional keys in the matrix/set.

        In the non-set expansion implementation, if the matrix does not yet
        exist (e.g., only ensemble specified), then the matrix is created.
        """
        # extract the ensemble key
        _ensemble = self.config_dict.pop('ensemble')

        # ensemble must be integer - check if valid
        if not isinstance(_ensemble, int):
            raise TypeError('Invalid ensemble type, must be an integer.')

        if _ensemble == 1:
            # there's nothing to do here
            warnings.warn(UserWarning(
                    'Ensemble was set to 1. '
                    'Remove this key from configuration.'))
            self._has_ensemble = False
            return

        if self._has_set:

            # pop the set to work with it, then replace at end
            old_set = self.config_dict.pop('set')
            n_ensembles = _ensemble  # from parsing above

            # here we augment each set by duplicating and add seed
            new_set = []  # new set list
            for ith_old_set in old_set:

                # loop through the number of ensembles
                for _ in range(n_ensembles):
                    # make a copy of the ith
                    jth_new_set = ith_old_set.copy()

                    # make a new seed for jth set
                    jth_seed = np.random.randint((2**32) - 1, dtype='u8')

                    # fill seed spec
                    jth_new_set['seed'] = jth_seed

                    # append to the new set list
                    new_set.append(jth_new_set)

            self.config_dict['set'] = new_set

        else:
            # if matrix does not exist, then it must be created
            if 'matrix' not in self.config_dict.keys():
                _matrix = {}
            else:
                _matrix = self.config_dict.pop('matrix')

            # check type of matrix
            #   is this needed here? Do earlier? Let it error naturally?
            if not isinstance(_matrix, dict):
                raise ValueError(
                    'Invalid matrix specification, '
                    'was not evaluated to "dict".')

            # check for invalid specs
            if 'seed' in _matrix.keys():
                raise ValueError(
                    'Random seeds cannot be specified in the matrix, '
                    'if an "ensemble" number is specified as well.')

            # generate list of random seeds
            seed_list = []
            n_ensembles = _ensemble
            for i in range(n_ensembles):
                seed_list.append(np.random.randint((2**32) - 1, dtype='u8'))

            # add list of random seeds to the matrix
            _matrix['seed'] = seed_list

            # write it back to the total configuration
            self.config_dict['matrix'] = _matrix

            # change the variable to expand matrix on future step
            self._has_matrix = True

    def _expand_set(self):

        # determine the set
        _set = self.config_dict.pop('set')

        # check that all sets are dictionaries
        if not isinstance(_set, list):
            raise TypeError(
                'Set list must be type `list` but was {}.'.format(type(_set)))
        for i, d in enumerate(_set):  # check validity of keys, depth == 1
            if not isinstance(d, dict):
                raise TypeError(
                    'Set must specify as a list of dictionaries')
            for k in d.keys():  # check validity of keys, depth == 1
                if ':' in k:
                    raise ValueError(
                        'Colon operator found in matrix expansion key.')

        # check that all sets have the same entries
        set0_set = set(_set[0].keys())
        for s in range(1, len(_set)):
            if not (set0_set == set(_set[s].keys())):
                raise ValueError('All keys in all sets must be identical.')

        # extract dimensionality of set
        jobs = len(_set)
        dims = len(_set[0])

        if self.verbose > 0:
            print(('Set expansion:\n' +
                   '  dims {_dims}\n' +
                   '  jobs {_jobs}').format(_dims=dims, _jobs=jobs))

        # preallocate the matrix expansion job yamls and output table
        self._matrix_table = np.empty(  # output table
            (jobs, dims+1), dtype='O')

        _fixed_config = self.config_dict.copy()  # fixed config dict

        # loop through each and create a config and add to _config_list
        for i in range(jobs):

            # being with the fixed config
            ith_config = _fixed_config.copy()

            # find job id and create output file
            ith_id = 'job_' + str(i).zfill(3)
            ith_dir = os.path.join(self._jobs_root, ith_id)

            # write the job number into output table
            self._matrix_table[i, 0] = ith_id

            # get config for this job
            ith_config['out_dir'] = ith_dir

            # loop through each var of this job
            for j, (key, val) in enumerate(_set[i].items()):

                # write info into fixed config dict
                ith_config[key] = val

                # write info into output table
                self._matrix_table[i, j+1] = val

            # add the configuration to a list to write out below
            self._config_list.append(ith_config)

        # store the matrix expansion
        #   this is useful for references by custom
        #   Python preprocessing / postprocessing
        matrix_table_file = os.path.join(
            self._jobs_root, 'jobs_parameters.txt')
        self._matrix_table_header = ', '.join(
            ['job_id', *[i for i in _set[0].keys()]])
        np.savetxt(matrix_table_file, self._matrix_table,
                   fmt='%s', delimiter=',', comments='',
                   header=self._matrix_table_header)

    def _expand_matrix(self):
        """Expand YAML matrix.

        Compute the matrix expansion of parameters listed in `matrix` key.
        """
        # extract and remove 'matrix' from config
        _matrix = self.config_dict.pop('matrix')

        # check validity of matrix specs
        if not isinstance(_matrix, dict):
            raise ValueError(
                'Invalid matrix specification, was not evaluated to "dict".')
        if 'out_dir' in _matrix.keys():
            raise ValueError(
                'You cannot specify "out_dir" as a matrix expansion key.')
        for k in _matrix.keys():  # check validity of keys, depth == 1
            if len(_matrix[k]) == 1:
                warnings.warn(UserWarning(
                    'Length of matrix key "%s" was 1, '
                    'remove this key from matrix configuration.' % str(k)))
            for v in _matrix[k]:
                if isinstance(v, list):
                    raise ValueError(
                        'Depth of matrix expansion must not be > 1')
            if ':' in k:
                raise ValueError(
                    'Colon operator found in matrix expansion key.')
            if k in self.config_dict.keys():
                raise ValueError(
                    'You cannot specify the same key in the matrix '
                    'configuration and fixed configuration. '
                    'Key "%s" was specified in both.' % str(k))

        # compute the expansion
        var_list = [k for k in _matrix.keys()]
        lil = [_matrix[v] for k, v in enumerate(var_list)]
        dims = len(lil)
        pts = [len(lst) for lst in lil]
        jobs = np.prod(pts)
        _combs = list(itertools.product(*lil))  # actual matrix expansion
        _fixed_config = self.config_dict.copy()  # fixed config dict

        if self.verbose > 0:
            print(('Matrix expansion:\n' +
                   '  dims {_dims}\n' +
                   '  jobs {_jobs}').format(_dims=dims, _jobs=jobs))

        # preallocate the matrix expansion job yamls and output table
        self._matrix_table = np.empty(  # output table
            (jobs, dims+1), dtype='O')

        # loop through each and create a config and add to _config_list
        for i in range(jobs):

            # begin with the fixed config
            ith_config = _fixed_config.copy()

            # find job id and create output file
            ith_id = 'job_' + str(i).zfill(3)
            ith_dir = os.path.join(self._jobs_root, ith_id)

            # write the job number into output table
            self._matrix_table[i, 0] = ith_id

            # get config for this job
            ith_config['out_dir'] = ith_dir

            # loop through each var of this job
            for j, val in enumerate(_combs[i]):

                # write info into fixed config dict
                ith_config[var_list[j]] = val

                # write info into output table
                self._matrix_table[i, j+1] = val

            # add the configuration to a list to write out below
            self._config_list.append(ith_config)

        # store the matrix expansion
        #   this is useful for references by custom
        #   Python preprocessing / postprocessing
        matrix_table_file = os.path.join(
            self._jobs_root, 'jobs_parameters.txt')
        self._matrix_table_header = ', '.join(['job_id', *var_list])
        np.savetxt(matrix_table_file, self._matrix_table,
                   fmt='%s', delimiter=',', comments='',
                   header=self._matrix_table_header)

    def _write_job_configs(self):

        if len(self._config_list) == 0:
            raise ValueError('Config list empty!')

        # loop through each job to write out info
        for c, config in enumerate(self.config_list):

            # write out the job specific yaml file
            # ith_p = self._write_yaml_config(c, config)
            if self.verbose > 0:
                print('Writing YAML file for job ' + str(int(c)))

            # ith_config = config['config']
            ith_dir = Path(config['out_dir'])       # job output folder
            ith_id = ith_dir.parts[-1]              # job id

            # create the output directory if needed
            if ith_dir.is_dir():
                if 'resume_checkpoint' in self.config_dict and \
                  self.config_dict['resume_checkpoint'] is True:
                    pass
                else:
                    raise FileExistsError(
                        'Job output directory (%s) already exists.' % str(ith_dir))
            else:
                ith_dir.mkdir()

            # write the file into the output directory
            ith_p = ith_dir / (str(ith_id) + '.yml')
            _write_yaml_config_to_file(config, ith_p)

            # append to the file list
            self._file_list.append(ith_p)

    def run_jobs(self, DeltaModel=None):
        """Run the set of jobs.

        This method can be seen as the actual execution stage of the
        preprocessor. If `--parallel` is specified in the command line or YAML
        file, the jobs will run in parallel.
        """
        if self._dryrun:
            return

        # process the special DeltaModel argument
        #  if not None, it is a class to be used by the Jobs
        if (DeltaModel is None):
            DeltaModel = BaseDeltaModel

        # initialize empty list to maintain reference to all Job instances
        num_total_processes = len(self.file_list)

        # NOTE: multiprocessing infrastructure is only available on linux.
        #       We only use parallel approach if the --parallel flag is given
        #       and we are running on linux.
        if 'parallel' in self.config_dict.keys():
            # and not (self.config_dict['parallel'] is None)):
            _parallel_flag = self.config_dict['parallel']
        else:
            _parallel_flag = False

        # if the parallel flag is given use the parallel infrastructure
        #   NOTE: the following evaluates to true for boolean `True` or
        #         any `int` > 0
        if _parallel_flag:
            # validate that os is Linux, otherwise error
            _os = platform.system()
            if _os != 'Linux':
                raise NotImplementedError(
                    'Parallel simulations only implemented on Linux.')

            # determine the number of processors to use
            if (_parallel_flag is True):
                # (number cores avail - 1), or 1 and never 0
                num_parallel_processes = (multiprocessing.cpu_count() - 1) or 1
            elif (isinstance(_parallel_flag, int)):
                num_parallel_processes = _parallel_flag
            else:
                raise ValueError('Parallel flag must be boolean or integer, '
                                 'but was {}, {}.'.format(type(_parallel_flag),
                                                          str(_parallel_flag)))
            # number of parallel processes is never greater than number of jobs
            num_parallel_processes = np.minimum(
                num_parallel_processes, num_total_processes)

            _msg = 'Running %g parallel jobs' % num_parallel_processes
            if self.verbose >= 1:
                print(_msg)

            # use Semaphore to limit number of concurrent Job
            s = multiprocessing.Semaphore(num_parallel_processes)

            # start a Queue for all Jobs
            q = multiprocessing.Queue()

            # loop and create and start all jobs
            for i in range(0, num_total_processes):
                s.acquire()  # aquire resource from Semaphore

                # open yaml file for specific job
                job_yaml = self._open_yaml(self.file_list[i])

                # apply job yaml to config dict
                #   enables complex job setups, where configs are edited
                #   between writing and .run_jobs()
                job_config = self.config_list[i].copy()
                job_config.update(job_yaml)

                # instantiate the job
                p = _ParallelJob(i=i, queue=q, sema=s,
                                 input_file=self.file_list[i],
                                 config_dict=job_config,
                                 DeltaModel=DeltaModel)
                self._job_list.append(p)
                p.start()

            # join processes to prevent ending the jobs before moving forward
            for i in self._job_list:
                i.join()

            # read from the queue and report (asynchronous...buggy...)
            time.sleep(1)
            while not q.empty():
                gotq = q.get()
                if gotq['code'] == 1:
                    print("Job {job} ended in error:\n {msg}".format_map(gotq))
                else:
                    print("Job {job} returned code {code} "
                          "for stage {stage}.".format_map(gotq))

        # if the parallel flag is False (default)
        elif (_parallel_flag is False):

            # loop and create all jobs
            for i in range(0, num_total_processes):

                # open yaml file for specific job
                job_yaml = self._open_yaml(self.file_list[i])

                # apply job yaml to config dict
                #   enables complex job setups, where configs are edited
                #   between writing and .run_jobs()
                job_config = self.config_list[i].copy()
                job_config.update(job_yaml)

                p = _SerialJob(i=i,
                               input_file=self.file_list[i],
                               config_dict=job_config,
                               DeltaModel=DeltaModel)
                self._job_list.append(p)

            # run the job(s)
            for i, job in enumerate(self._job_list):
                if self.verbose > 0:
                    print("Starting job %s" % str(i))
                job.run()

        # if the parallel flag is a junk value
        else:
            raise ValueError

        self._is_completed = True


class _BaseJob(abc.ABC):
    """Base class for individual jobs to run via the preprocessor.

    The base class handles setting options for the run time duration, based on
    the inputs to the command line and the yaml file.

    .. important:: the :meth:`run` method must be implemented by subclasses.

    .. note:: You probably don't need to interact with this class directly.
    """

    def __init__(self, i, input_file, config_dict,
                 DeltaModel=None, defer_output=False):
        """Initialize a job.

        The `input_file` argument is passed to the DeltaModel for
        instantiation.

        The various model run duration parameters are passed from the
        `cli_dict` and `yaml_dict` arguments, and are processed into a
        single value for the run time. Precedence is given to values
        specified in the command line interface.
        """
        self.i = i
        self.input_file = input_file

        # if 'DeltaModel' in config_dict.keys():
        #     _DM = config_dict['DeltaModel']
        # else:
        #     _DM = DeltaModel
        if (DeltaModel is None):
            DeltaModel = BaseDeltaModel

        self.deltamodel = DeltaModel(input_file=input_file,
                                     defer_output=defer_output)
        _curr_time = self.deltamodel._time

        # process If
        if 'If' in config_dict.keys():
            self._If = float(config_dict['If'])
        else:
            self._If = 1.0

        # determine job end time, *in model time*
        if ('timesteps' in config_dict.keys()):
            # fill informational fields
            self._time_type = 'timesteps'
            self._time_config = config_dict['timesteps']

            # compute the end time
            self._job_end_time = _curr_time + \
                ((config_dict['timesteps'] * self.deltamodel._dt))

        elif ('time' in config_dict.keys()):
            # fill informational fields
            self._time_type = 'time'
            self._time_config = config_dict['time']

            # compute the end time
            self._job_end_time = _curr_time + \
                (config_dict['time'] * self._If)

        elif ('time_years' in config_dict.keys()):
            # fill informational fields
            self._time_type = 'time_years'
            self._time_config = config_dict['time_years']

            # compute the end time
            self._job_end_time = _curr_time + \
                (config_dict['time_years'] * self._If * 86400 * 365.25)
        else:
            raise ValueError(
                'You must specify a run duration configuration in either '
                'the input YAML file or via input arguments.')

    @abc.abstractmethod
    def run(self):
        """Implementation to run the jobs as needed.

        .. note::

            Consider using a series of ``try-except-else-finally` statements to
            ensure that this method handles all outcomes of the simulations.
        """
        ...


class _SerialJob(_BaseJob):
    """Serial job run by the preprocessor.

    This class is a subclass of :obj:`_BaseJob` and implements the :meth:`run`
    method for jobs that should occur in serial.

    .. note:: You probably don't need to interact with this class directly.
    """

    def __init__(self, i, input_file, config_dict, DeltaModel=None,):
        """Initialize a serial job.

        The `input_file` argument is passed to the DeltaModel for
        instantiation.

        The various model run duration parameters are passed from the
        `config_dict` argument, and are processed into a single value for the
        run time.
        """
        super().__init__(i, input_file, config_dict,
                         DeltaModel=DeltaModel, defer_output=False)

    def run(self):
        """Loop the model.

        Iterate the timestep ``update`` routine for the specified number of
        iterations.
        """
        # try to initialize and run the model
        try:
            # run the simualtion
            while self.deltamodel._time < self._job_end_time:
                self.deltamodel.update()

        # if the model run fails
        except (RuntimeError, ValueError) as e:
            _msg = ', '.join(['job: ' + str(self.i), 'stage: ' + '1',
                             'code: ' + '1', 'msg: ' + str(e)])
            self.deltamodel.logger.error(_msg)
            self.deltamodel.logger.exception(e)

        # if the model run succeeds
        else:
            _msg = ', '.join(['job: ' + str(self.i), 'stage: ' + '1',
                             'code: ' + '0'])
            self.deltamodel.logger.info(_msg)

        # try to finalize the model
        try:
            self.deltamodel.finalize()

        # if the model finalization fails
        except (RuntimeError, ValueError) as e:
            _msg = ', '.join(['job: ' + str(self.i), 'stage: ' + '2',
                             'code: ' + '1', 'msg: ' + str(e)])
            self.deltamodel.logger.error(_msg)
            self.deltamodel.logger.exception(e)

        # if the finalization succeeds
        else:
            _msg = ', '.join(['job: ' + str(self.i), 'stage: ' + '2',
                             'code: ' + '0'])
            self.deltamodel.logger.info(_msg)


class _ParallelJob(_BaseJob, multiprocessing.Process):
    """Parallel job run by the preprocessor.

    This class is a subclass of :obj:`_BaseJob` and implements the :meth:`run`
    method for jobs that should occur in parallel.

    .. note:: You probably don't need to interact with this class directly.
    """

    def __init__(self, i, queue, sema,
                 input_file, config_dict, DeltaModel=None):
        """Initialize a parallel job.

        The `input_file` argument is passed to the DeltaModel for
        instantiation.

        The various model run duration parameters are passed from the
        `config_dict` argument, and are processed into a single value for the
        run time.
        """
        # inherit with explicit method resolution order
        _BaseJob.__init__(self, i, input_file, config_dict,
                          DeltaModel=DeltaModel, defer_output=True)
        multiprocessing.Process.__init__(self)

        self.queue = queue
        self.sema = sema  # semaphore for limited multiprocessing

    def run(self):
        """Run the model, with infrastructure for parallel.

        Iterate the timestep ``update`` routine for the specified number of
        iterations.
        """
        self.queue.put({'job': self.i, 'stage': 0, 'code': 0})

        # overall wrapped in try to ensure sema is safely released
        try:
            # try to initialize and run the model
            try:
                # initialize the output files (defer_output=True above)
                # unless resuming from checkpoint, then load the checkpoint
                if self.deltamodel.resume_checkpoint:
                    # here we set defer output to false when loading the
                    #   checkpoint on this thread
                    self.deltamodel.load_checkpoint(defer_output=False)
                else:
                    # infrastructure deferred, need to trigger manually
                    self.deltamodel.hook_init_output_file()
                    self.deltamodel.init_output_file()

                    self.deltamodel.hook_output_data()
                    self.deltamodel.output_data()

                    self.deltamodel.hook_output_checkpoint()
                    self.deltamodel.output_checkpoint()

                # run the simualtion
                while self.deltamodel._time < self._job_end_time:
                    self.deltamodel.update()

            # if the model run fails
            except (RuntimeError, ValueError) as e:
                self.queue.put({'job': self.i, 'stage': 1,
                                'code': 1, 'msg': str(e)})
                self.deltamodel.logger.error(str(e))
                self.deltamodel.logger.exception(e)

            # if the model run succeeds
            else:
                self.queue.put({'job': self.i, 'stage': 1,
                                'code': 0})

            # try to finalize the model
            try:
                self.deltamodel.finalize()

            # if the model finalization fails
            except (RuntimeError, ValueError) as e:
                self.queue.put({'job': self.i, 'stage': 2,
                                'code': 1, 'msg': str(e)})
                self.deltamodel.logger.error(str(e))
                self.deltamodel.logger.exception(e)

            # if the finalization succeeds
            else:
                self.queue.put({'job': self.i, 'stage': 2,
                                'code': 0})

        # ALWAYS release the semaphore, so other runs can continue
        finally:
            self.sema.release()


class PreprocessorCLI(BasePreprocessor):
    """Command line preprocessor.

    This is the main CLI class that is called from the command line. The class
    defines a method to process the arguments from the command line (using the
    `argparse` package).

    .. note::

        You probably do not need to interact with this class directly.
        Instead, you can use the command line API as it is described in the
        :doc:`User Guide </guides/user_guide>` or the python API
        :class:`~pyDeltaRCM.preprocessor.Preprocessor`.

        When the class is called from the command line the instantiated
        object's method :meth:`run_jobs` is called by the
        :obj:`~pyDeltaRCM.preprocessor.preprocessor_wrapper` function that the
        CLI (entry point) calls directly.

    """

    def __init__(self):
        """Initialize the CLI preprocessor.

        The initialization includes the entire configuration of the job list
        (parsing, timesteps, etc.). The jobs are *not* run automatically
        during instantiation of the class.
        """
        super().__init__()

        # process command line to a dictionary
        cli_dict = self.process_arguments()

        # process the input file to a dictionary (or empty if none)
        if cli_dict['config'] is not None:
            input_file = cli_dict['config']
            yaml_dict = self.open_input_file_to_dict(input_file)
            self._input_file = input_file  # fill field
        else:
            yaml_dict = {}

        # combine the dicts into a single config
        self.config_dict = {}
        self.config_dict.update(yaml_dict)
        self.config_dict.update(cli_dict)

        # construct file list (expansions)
        self.construct_file_list()

    def process_arguments(self):
        """Process the command line arguments.

        .. note::

            The command line args are not directly passed to this function in
            any way.

        """
        parser = argparse.ArgumentParser(
            description='Arguments for running pyDeltaRCM from command line')

        parser.add_argument(
            '--config', help='Path to a config file that you would like '
            'to use.')
        parser.add_argument(
            '--timesteps', help='Number of timesteps to run model. Optional, '
            'and if provided, will override any value in the config file. '
            'Is overrided by "time" options.')
        parser.add_argument(
            '--time', help='Time to run the model until, in seconds.'
            'Time is scaled due to intermittency, which is set to 1 by '
            'default (i.e., no scaling). See options "If" or "time_years" '
            'for other specification options. Optional, '
            'and if provided, will override any value in the config file.')
        parser.add_argument(
            '--time_years', help='Time to run the model until, in years.'
            'Time is scaled due to intermittency, which is set to 1 by '
            'default (i.e., no scaling). See options "If" or "time" '
            'for other specification options. Optional, '
            'and if provided, will override any value in the config file.')
        parser.add_argument(
            '--If', help='Assumed intermittency factor for converting model '
            'time to real-world time. This is a value in (0, 1] representing '
            'the fraction of days per year that the delta is flooding. '
            'Optional, default value is 1.')
        parser.add_argument(
            '--dryrun', action='store_true',
            help='Boolean indicating whether to execute timestepping or only '
            'set up the run.')
        parser.add_argument(
            '--parallel', type=int, nargs='?', const=True,
            help='Run jobs in parallel, if possible. If given without any'
            'argument, the number of parallel cores used depends '
            'on system specs as (ncores - 1). Specify an integer value to '
            'specify the number of cores to be used. '
            'Optional, default value is False.')
        parser.add_argument(
            '--version', action='version',
            version=_ver, help='Print pyDeltaRCM version.')

        # parse and convert to dict
        args = parser.parse_args()
        args_dict = vars(args)

        # convert arguments to valid types for preprocessor
        if not (args_dict['timesteps'] is None):
            args_dict['timesteps'] = int(args_dict['timesteps'])
        else:
            args_dict.pop('timesteps')
        if not (args_dict['time'] is None):
            args_dict['time'] = float(args_dict['time'])
        else:
            args_dict.pop('time')
        if not (args_dict['time_years'] is None):
            args_dict['time_years'] = float(args_dict['time_years'])
        else:
            args_dict.pop('time_years')

        # set (or remove) defaults as needed
        if (args_dict['parallel'] is None):
            # if not given remove the key from cli spec
            args_dict.pop('parallel')
        if (args_dict['If'] is None):
            # if not given remove the key from cli spec
            args_dict.pop('If')

        return args_dict


class Preprocessor(BasePreprocessor):
    """Python high level api.

    This is the python high-level API class that is callable from a python
    script. For complete documentation on the API configurations, see the
    :doc:`User Guide </guides/user_guide>`.

    The class gives a way to configure and run multiple jobs from a python
    script.

    Examples
    --------

    To configure a set of jobs (or a single job), instantiate the python
    preprocessor and then manually run the configured jobs:

    .. code::

        >>> pp = preprocessor.Preprocessor(input_file=p, timesteps=500)
        >>> pp.run_jobs()

    """

    def __init__(self, input_file=None, **kwargs):
        """Initialize the python preprocessor.

        The initialization includes the entire configuration of the job list
        (parsing, timesteps, etc.). The jobs are *not* run automatically
        during instantiation of the class.

        You must specify timesteps in either the YAML configuration file or
        via the `timesteps` parameter.

        Parameters
        ----------
        input_file : :obj:`str`, optional
            Path to an input YAML configuration file. Must include the
            `timesteps` parameter if you do not specify the `timesteps` as a
            keyword argument.

        timesteps : :obj:`int`, optional
            Number of timesteps to run each of the jobs. Must be specified if
            you do not specify the `timesteps` parameter in the input YAML
            file.

        """
        super().__init__()

        if not input_file and len(kwargs.keys()) == 0:
            raise ValueError('Cannot use Preprocessor with no arguments.')

        # process "command line" to a dict (already is)
        cli_dict = kwargs

        # process the input yaml file to a dict
        if input_file is None:
            yaml_dict = {}
        else:
            yaml_dict = self.open_input_file_to_dict(input_file)
            self._input_file = input_file  # fill field

        # combine the dicts into a single config
        self.config_dict = {}
        self.config_dict.update(yaml_dict)
        self.config_dict.update(cli_dict)

        # construct file list (expansions)
        self.construct_file_list()


def preprocessor_wrapper():
    """Wrapper for command line interface.

    The `entry_points` setup of a command line interface requires a function,
    so we use this simple wrapper to instantiate and run the jobs.

    This function creates an instance of the
    :obj:`~pyDeltaRCM.preprocessor.PreprocessorCLI` and calls the
    :meth:`~pyDeltaRCM.preprocessor.PreprocessorCLI.run_jobs` to execute all
    jobs configured in the preprocessor. In code:

    .. code:: python

        pp = PreprocessorCLI()
        pp.run_jobs()
    """
    pp = PreprocessorCLI()
    pp.run_jobs()


def _write_yaml_config_to_file(_config, _path):
    """Write a config to file in output folder.

    Write the entire yaml configuation for the configured job out to a
    file in the job output foler.

    .. note::

        This function is utilized by the BMI implementation of pyDeltaRCM as
        well. Please do not move.
    """
    def _write_parameter_to_file(f, varname, varvalue):
        """Write each line, formatted."""
        if varvalue is None:
            f.write(varname + ': ' + 'null' + '\n')
        else:
            f.write(varname + ': ' + str(varvalue) + '\n')

    f = open(_path, "a")
    for k in _config.keys():
        _write_parameter_to_file(f, k, _config[k])
    f.close()


def scale_relative_sea_level_rise_rate(mmyr, If=1):
    """Scale a relative sea level rise rate to model time.

    This function scales any relative sea level rise rate (RSLR) (e.g., sea
    level rise, subsidence) to a rate appropriate for the model time. This is
    helpful, because most discussion of RSLR uses units of mm/yr, but the
    model (and model configuration) require units of m/s. Additionally, the
    model framework needs to assume an "intermittency factor" to convert from
    real-world time to model time.

    Relative sea level rise (subsidence and/or sea level rise) are scaled from
    real world dimensions of mm/yr to model input as:

    .. math::

        \\widehat{RSLR} = (RSLR / 1000) \\cdot \\dfrac{1}{I_f \\cdot 365.25 \\cdot 86400}

    This conversion makes it such that when one real-world year has elapsed
    (:math:`I_f \\cdot 365.25 \\cdot 86400` seconds in model time), the relative
    sea level has changed by the number of millimeters specified in the input
    :obj:`mmyr`.

    .. note::

        Users should use this function to determine the value to specify in
        an input YAML configuration file; no scaling is performed
        internally.

    Parameters
    ----------
    mmyr : :obj:`float`
        Millimeters per year, relative sea level rise rate.

    If : :obj:`float`, optional
        Intermittency factor, fraction of time represented by morphodynamic
        activity. Should be in interval (0, 1). Defaults to 1 if not provided,
        i.e., no scaling is performed.

    Returns
    -------
    scaled : :obj:`float`
        Scaled relative sea level rise rate, in meters per second.
    """
    return (mmyr / 1000) * (1 / (shared_tools._scale_factor(
        If, units='years')))


# make the connection for running the preprocessor directly
if __name__ == '__main__':

    preprocessor_wrapper()



# unit tests for consistent model outputs

import pytest

import os
import platform
import shutil
import numpy as np
from netCDF4 import Dataset

import unittest.mock as mock

from pyDeltaRCM import DeltaModel
from pyDeltaRCM import preprocessor

from .. import utilities


@mock.patch(
    'pyDeltaRCM.iteration_tools.iteration_tools.solve_water_and_sediment_timestep',
    new=utilities.FastIteratingDeltaModel.solve_water_and_sediment_timestep)
class TestCheckpointingIntegrations:
    """
    The above patch implements an augmented DeltaModel from `utilities`. In
    this modified DeltaModel, the `solve_water_and_sediment_timestep`
    operations (i.e., the time consuming part of the model) is replaced with
    an updating random field. This guarantees that the random-repeatedness of
    checkpointing is validated, but it is much faster and easier to isolate
    checkpointing-related issues from model issues.
    """

    def test_simple_checkpoint(self, tmp_path):
        """Test checkpoint vs a base run.

        Also, checks resumed model against another checkpoint run.
        """
        # define a yaml for the longer model run
        file_name = 'base_run.yaml'
        base_p, base_f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(base_f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(base_f, 'save_checkpoint', True)
        base_f.close()
        longModel = DeltaModel(input_file=base_p)

        # run for some number of updates
        for _ in range(0, 50):
            longModel.update()
        longModel.finalize()

        # try defining a new model but plan to load checkpoint from longModel
        file_name = 'base_run.yaml'
        base_p, base_f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(base_f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(base_f, 'resume_checkpoint', True)
        base_f.close()
        resumeModel = DeltaModel(input_file=base_p)

        # advance the resumed model until it catch up to longModel
        assert resumeModel.time < longModel.time
        while resumeModel._time < longModel._time:
            resumeModel.update()
        resumeModel.finalize()

        # the longModel and resumeModel should match
        assert longModel.time == resumeModel.time
        assert np.all(longModel.eta == resumeModel.eta)
        assert np.all(longModel.uw == resumeModel.uw)
        assert np.all(longModel.ux == resumeModel.ux)
        assert np.all(longModel.uy == resumeModel.uy)
        assert np.all(longModel.depth == resumeModel.depth)
        assert np.all(longModel.stage == resumeModel.stage)
        assert np.all(longModel.sand_frac == resumeModel.sand_frac)
        assert np.all(longModel.active_layer == resumeModel.active_layer)

        # define another model that loads the checkpoint
        file_name = 'base_run.yaml'
        base_p, base_f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(base_f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(base_f, 'resume_checkpoint', True)
        base_f.close()
        resumeModel2 = DeltaModel(input_file=base_p)

        # advance the resumed model until it catch up to longModel
        while resumeModel2._time < resumeModel._time:
            resumeModel2.update()
        resumeModel2.finalize()

        # the two models that resumed from the checkpoint should be the same
        assert resumeModel2.time == resumeModel.time
        assert np.all(resumeModel2.uw == resumeModel.uw)
        assert np.all(resumeModel2.ux == resumeModel.ux)
        assert np.all(resumeModel2.uy == resumeModel.uy)
        assert np.all(resumeModel2.depth == resumeModel.depth)
        assert np.all(resumeModel2.stage == resumeModel.stage)
        assert np.all(resumeModel2.sand_frac == resumeModel.sand_frac)
        assert np.all(resumeModel2.active_layer == resumeModel.active_layer)

    def test_checkpoint_nc(self, tmp_path):
        """Test the netCDF that is written to by the checkpointing."""
        # define a yaml for the base model run
        file_name = 'base_run.yaml'
        base_p, base_f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(base_f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(base_f, 'save_eta_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_depth_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_discharge_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_sandfrac_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_checkpoint', True)
        base_f.close()
        baseModel = DeltaModel(input_file=base_p)

        # run for some base number of steps
        nt_base = 50
        for _ in range(0, 50):
            baseModel.update()

        # force the model run to end immmediately after exporting a checkpoint
        nt_var = 0
        while (baseModel._save_time_since_checkpoint != 0):
            baseModel.update()
            nt_var += 1

        # then finalize
        baseModel.finalize()

        # check that the time makes sense
        assert baseModel.time == baseModel._dt * (nt_base + nt_var)

        # try defining a new model but plan to load checkpoint from baseModel
        file_name = 'base_run.yaml'
        base_p, base_f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(base_f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(base_f, 'save_eta_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_depth_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_discharge_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_sandfrac_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_checkpoint', False)
        utilities.write_parameter_to_file(base_f, 'resume_checkpoint', True)
        base_f.close()
        resumeModel = DeltaModel(input_file=base_p)

        assert resumeModel.time == baseModel.time  # same when resumed

        # advance it until output_data has been called again
        nt_resume = 0
        while (resumeModel._save_time_since_data != 0) or (nt_resume < 50):
            resumeModel.update()
            nt_resume += 1
        resumeModel.finalize()

        assert nt_resume > 0
        assert resumeModel.time > baseModel.time

        # assert that output netCDF4 exists
        exp_path_nc = os.path.join(tmp_path / 'test', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        # load it into memory and check values in the netCDF4
        output = Dataset(exp_path_nc, 'r', allow_pickle=True)
        out_vars = output.variables.keys()

        # check that expected variables are in the file
        assert 'x' in out_vars
        assert 'y' in out_vars
        assert 'time' in out_vars
        assert 'eta' in out_vars
        assert 'depth' in out_vars
        assert 'discharge' in out_vars
        assert 'sandfrac' in out_vars

        # check attributes of variables
        assert output['time'][0].tolist() == 0.0
        assert output['time'][-1] == resumeModel.time
        assert output['time'][-1].tolist() == resumeModel._dt * \
            (nt_base + nt_var + nt_resume)
        assert output['eta'][0].shape == resumeModel.eta.shape
        assert output['eta'][-1].shape == resumeModel.eta.shape
        assert output['depth'][-1].shape == resumeModel.eta.shape
        assert output['discharge'][-1].shape == resumeModel.eta.shape
        assert output['sandfrac'][-1].shape == resumeModel.eta.shape
        # check the metadata
        assert output['meta']['L0'][:] == resumeModel.L0
        assert output['meta']['N0'][:] == resumeModel.N0
        assert output['meta']['CTR'][:] == resumeModel.CTR
        assert output['meta']['dx'][:] == resumeModel.dx
        assert output['meta']['h0'][:] == resumeModel.h0
        assert np.all(output['meta']['cell_type'][:] == resumeModel.cell_type)
        assert output['meta']['H_SL'][-1].data == resumeModel.H_SL
        assert output['meta']['f_bedload'][-1].data == resumeModel.f_bedload
        C0_from_file = float(output['meta']['C0_percent'][-1].data)
        assert pytest.approx(C0_from_file) == resumeModel.C0_percent
        assert output['meta']['u0'][-1].data == resumeModel.u0

        # checkpoint interval aligns w/ timestep dt so these should match
        assert output['time'][-1].tolist() == resumeModel.time

    def test_checkpoint_diff_dt(self, tmp_path):
        """Test when checkpoint_dt does not match dt or save_dt."""
        # define a yaml for the base model run
        file_name = 'base_run.yaml'
        base_p, base_f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(base_f, 'save_eta_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_depth_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_discharge_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_checkpoint', True)
        utilities.write_parameter_to_file(base_f, 'out_dir', tmp_path / 'test')
        base_f.close()
        baseModel = DeltaModel(input_file=base_p)

        # modify the checkpoint dt to be different than save_dt
        baseModel._checkpoint_dt = (baseModel.save_dt * 0.65)

        for _ in range(0, 50):
            baseModel.update()
        baseModel.finalize()

        assert baseModel.time == baseModel._dt * 50
        baseModelSavedTime = (baseModel.time -
                              baseModel._save_time_since_checkpoint)
        assert baseModelSavedTime > 0

        # try defining a new model but plan to load checkpoint from baseModel
        file_name = 'base_run.yaml'
        base_p, base_f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(base_f, 'save_eta_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_depth_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_discharge_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_checkpoint', False)
        utilities.write_parameter_to_file(base_f, 'resume_checkpoint', True)
        utilities.write_parameter_to_file(base_f, 'out_dir', tmp_path / 'test')
        base_f.close()
        resumeModel = DeltaModel(input_file=base_p)

        assert resumeModel.time == baseModelSavedTime

        # advance until some steps and just saved
        nt_resume = 0
        while (resumeModel._save_time_since_data != 0) or (nt_resume < 50):
            resumeModel.update()
            nt_resume += 1
        resumeModel.finalize()

        # assert that output netCDF4 exists
        exp_path_nc = os.path.join(tmp_path / 'test', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        # load it into memory and check values in the netCDF4
        output = Dataset(exp_path_nc, 'r', allow_pickle=True)
        out_vars = output.variables.keys()
        # check that expected variables are in the file
        assert 'x' in out_vars
        assert 'y' in out_vars
        assert 'time' in out_vars
        assert 'eta' in out_vars
        assert 'depth' in out_vars
        assert 'discharge' in out_vars
        # check attributes of variables
        assert output['time'][0].tolist() == 0.0
        assert output['time'][-1].tolist() == resumeModel.time

    def test_multi_checkpoints(self, tmp_path):
        """Test using checkpoints multiple times for a given model run."""
        # define a yaml for the base model run
        file_name = 'base_run.yaml'
        base_p, base_f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(base_f, 'save_eta_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_checkpoint', True)
        utilities.write_parameter_to_file(base_f, 'out_dir', tmp_path / 'test')
        base_f.close()
        baseModel = DeltaModel(input_file=base_p)

        # run base for 2 timesteps
        for _ in range(0, 50):
            baseModel.update()
        baseModel.finalize()

        # try defining a new model but plan to load checkpoint from baseModel
        file_name = 'base_run.yaml'
        base_p, base_f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(base_f, 'save_eta_grids', True)
        utilities.write_parameter_to_file(base_f, 'save_checkpoint', True)
        utilities.write_parameter_to_file(base_f, 'resume_checkpoint', True)
        utilities.write_parameter_to_file(base_f, 'out_dir', tmp_path / 'test')
        base_f.close()
        resumeModel = DeltaModel(input_file=base_p)

        assert resumeModel.time <= baseModel.time

        # advance it more steps
        for _ in range(0, 25):
            resumeModel.update()
        resumeModel.finalize()

        # create another resume model
        resumeModel02 = DeltaModel(input_file=base_p)

        assert resumeModel02.time <= resumeModel.time  # should be same

        # step it some more
        nt_resume02 = 0
        while (resumeModel02._save_time_since_data != 0) or (nt_resume02 < 50):
            resumeModel02.update()
            nt_resume02 += 1

        # assert that output netCDF4 exists
        exp_path_nc = os.path.join(tmp_path / 'test', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        # load it into memory and check values in the netCDF4
        output = Dataset(exp_path_nc, 'r', allow_pickle=True)
        out_vars = output.variables.keys()
        # check that expected variables are in the file
        assert 'x' in out_vars
        assert 'y' in out_vars
        assert 'time' in out_vars
        assert 'eta' in out_vars
        # check attributes of variables
        assert output['time'][0].tolist() == 0.0
        assert output['time'][-1].tolist() == resumeModel02.time

    def test_load_nocheckpoint(self, tmp_path):
        """Try loading a checkpoint file when one doesn't exist."""
        # define a yaml
        file_name = 'trial_run.yaml'
        base_p, base_f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(base_f, 'resume_checkpoint', True)
        utilities.write_parameter_to_file(base_f, 'out_dir', tmp_path / 'test')
        base_f.close()

        # try loading the model yaml despite no checkpoint existing
        with pytest.raises(FileNotFoundError):
            _ = DeltaModel(input_file=base_p)

    @pytest.mark.skipif(
        platform.system() != 'Linux',
        reason='Parallel support only on Linux OS.')
    def test_py_hlvl_parallel_checkpoint(self, tmp_path):
        """Test checkpointing in parallel."""
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'ensemble', 2)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(f, 'parallel', 2)
        utilities.write_parameter_to_file(f, 'save_checkpoint', True)
        utilities.write_parameter_to_file(f, 'save_eta_grids', True)
        f.close()
        pp = preprocessor.Preprocessor(input_file=p, timesteps=50)
        # assertions for job creation
        assert len(pp.file_list) == 2
        assert pp._is_completed is False

        # run the jobs, mocked deltas
        pp.run_jobs()

        # compute the expected final time recorded
        _dt = pp.job_list[1].deltamodel._dt
        _checkpoint_dt = pp.job_list[1].deltamodel._checkpoint_dt
        expected_save_interval = (((_checkpoint_dt // _dt) + 1) * _dt)
        expected_last_save_time = (((50 * _dt) // expected_save_interval) *
                                   expected_save_interval)

        # assertions after running jobs
        assert isinstance(pp.job_list[0], preprocessor._ParallelJob)
        assert pp._is_completed is True
        exp_path_nc0 = os.path.join(
            tmp_path / 'test', 'job_000', 'pyDeltaRCM_output.nc')
        exp_path_nc1 = os.path.join(
            tmp_path / 'test', 'job_001', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc0)
        assert os.path.isfile(exp_path_nc1)
        # check that checkpoint files exist
        exp_path_ckpt0 = os.path.join(
            tmp_path / 'test', 'job_000', 'checkpoint.npz')
        exp_path_ckpt1 = os.path.join(
            tmp_path / 'test', 'job_001', 'checkpoint.npz')
        assert os.path.isfile(exp_path_ckpt0)
        assert os.path.isfile(exp_path_ckpt1)
        # load one output files and check values
        out_old = Dataset(exp_path_nc1)
        assert 'meta' in out_old.groups.keys()
        assert out_old['time'][0].tolist() == 0.0
        assert out_old['time'][-1].tolist() == expected_last_save_time

        # close netCDF file
        out_old.close()

        # try to resume jobs
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'ensemble', 2)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(f, 'parallel', 2)
        utilities.write_parameter_to_file(f, 'resume_checkpoint', True)
        utilities.write_parameter_to_file(f, 'save_eta_grids', True)
        f.close()
        pp = preprocessor.Preprocessor(input_file=p, timesteps=50)
        # assertions for job creation
        assert len(pp.file_list) == 2
        assert pp._is_completed is False

        # run the jobs, mocked deltas
        pp.run_jobs()

        # assertions after running jobs
        assert isinstance(pp.job_list[0], preprocessor._ParallelJob)
        assert pp._is_completed is True
        exp_path_nc0 = os.path.join(
            tmp_path / 'test', 'job_000', 'pyDeltaRCM_output.nc')
        exp_path_nc1 = os.path.join(
            tmp_path / 'test', 'job_001', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc0)
        assert os.path.isfile(exp_path_nc1)
        # check that checkpoint files still exist
        exp_path_ckpt0 = os.path.join(
            tmp_path / 'test', 'job_000', 'checkpoint.npz')
        exp_path_ckpt1 = os.path.join(
            tmp_path / 'test', 'job_001', 'checkpoint.npz')
        assert os.path.isfile(exp_path_ckpt0)
        assert os.path.isfile(exp_path_ckpt1)
        # load one output file to check it out
        out_fin = Dataset(exp_path_nc1)
        assert 'meta' in out_old.groups.keys()
        assert out_fin['time'][0].tolist() == 0
        assert out_fin['time'][-1].tolist() == expected_last_save_time * 2
        # close netcdf file
        out_fin.close()


class TestCheckpointingCreatingLoading:

    def test_load_checkpoint_with_netcdf(self, tmp_path):
        """Test that a run can be resumed when there are outputs.
        """
        # define a yaml with outputs (defaults will output strata)
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True,
                                      'save_eta_grids': True})
        _delta = DeltaModel(input_file=p)

        # replace eta with a random field for checkpointing success check
        _rand_field = np.random.uniform(0, 1, size=_delta.eta.shape)
        _delta.eta = _rand_field

        _delta._save_time_since_checkpoint = float("inf")
        _delta.output_checkpoint()  # force another checkpoint
        _delta.finalize()

        # paths exists
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'pyDeltaRCM_output.nc'))
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'checkpoint.npz'))
        _delta = []  # clear

        # can be resumed
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True,
                                      'resume_checkpoint': True})
        _delta = DeltaModel(input_file=p)

        # check that fields match
        assert np.all(_delta.eta == _rand_field)

    def test_create_checkpoint_without_netcdf(self, tmp_path):
        """Test that a checkpoint can be created when there are no outputs
        """
        # define a yaml with NO outputs, but checkpoint
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True})

        _delta = DeltaModel(input_file=p)

        # replace eta with a random field for checkpointing success check
        _rand_field = np.random.uniform(0, 1, size=_delta.eta.shape)
        _delta.eta = _rand_field

        _delta._save_time_since_checkpoint = float("inf")
        _delta.output_checkpoint()  # force another checkpoint
        _delta.finalize()

        # should be no file
        assert not os.path.isfile(os.path.join(
            _delta.prefix, 'pyDeltaRCM_output.nc'))

        # can be resumed
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True,
                                      'resume_checkpoint': True})
        _delta = DeltaModel(input_file=p)

        # check that fields match
        assert np.all(_delta.eta == _rand_field)

    def test_load_checkpoint_without_netcdf(self, tmp_path):
        """Test that a run can be resumed when there are outputs.
        """
        # define a yaml with NO outputs, but checkpoint
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True})
        _delta = DeltaModel(input_file=p)

        # replace eta with a random field for checkpointing success check
        _rand_field = np.random.uniform(0, 1, size=_delta.eta.shape)
        _delta.eta = _rand_field

        _delta._save_time_since_checkpoint = float("inf")
        _delta.output_checkpoint()  # force another checkpoint
        _delta.finalize()

        # should be no nc file but should be a checkpoint file
        assert not os.path.isfile(os.path.join(
            _delta.prefix, 'pyDeltaRCM_output.nc'))
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'checkpoint.npz'))

        # now try to resume, will WARN on not finding netcdf
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True,
                                      'save_eta_grids': True,
                                      'resume_checkpoint': True})
        with pytest.warns(UserWarning, match=r'NetCDF4 output *.'):
            _delta = DeltaModel(input_file=p)

        # assert that a new output file exists file exists
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'pyDeltaRCM_output.nc'))
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'checkpoint.npz'))

        # check that fields match
        assert np.all(_delta.eta == _rand_field)

    @mock.patch(
        'pyDeltaRCM.iteration_tools.iteration_tools.solve_water_and_sediment_timestep',
        new=utilities.FastIteratingDeltaModel.solve_water_and_sediment_timestep)
    @pytest.mark.skipif(
        platform.system() != 'Linux',
        reason='Parallel support only on Linux OS.')
    def test_load_ckpt_wo_netcdf_parallel_spinup_to_matrix(self, tmp_path):
        """
        Test that multiple matrix runs can be resumed from a single checkpoint
        file, and take advantage of the preprocessor parallel infrastructure.
        """
        # define a yaml with an output and checkpoint
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True,
                                      'save_eta_grids': True})
        baseModel = DeltaModel(input_file=p)

        # run base for 2 timesteps
        for _ in range(0, 50):
            baseModel.update()
        baseModel.finalize()

        # get the number of times the data has been output in the base file
        bmsi = baseModel._save_iter
        assert bmsi > 0

        # check that files exist, and then delete nc
        assert os.path.isfile(os.path.join(
            baseModel.prefix, 'pyDeltaRCM_output.nc'))
        assert os.path.isfile(os.path.join(
            baseModel.prefix, 'checkpoint.npz'))

        # open the file and check dimensions
        exp_nc_path = os.path.join(baseModel.prefix, 'pyDeltaRCM_output.nc')
        output = Dataset(exp_nc_path, 'r', allow_pickle=True)
        out_vars = output.variables.keys()
        # check that expected variables are in the file
        assert 'x' in out_vars
        assert 'y' in out_vars
        assert 'time' in out_vars
        assert 'eta' in out_vars
        # check attributes of variables
        assert output['time'].shape[0] == bmsi

        ########################

        # set up a matrix of runs
        resume_dict = {'save_checkpoint': False,
                       'resume_checkpoint': True,
                       'save_eta_grids': True,
                       'out_dir': os.path.join(tmp_path, 'matrix'),
                       'parallel': 4}
        _matrix = {'f_bedload': [0.1, 0.2, 0.5, 1]}
        resume_dict['matrix'] = _matrix

        # let the preprocessor write the initial matrix and
        #    create a new output netcdf file
        pp = preprocessor.Preprocessor(
            resume_dict,
            timesteps=25)  # 2000

        # now copy the checkpoint
        for j, j_file in enumerate(pp.file_list):
            # get folder
            j_folder = j_file.parent

            # copy the spinup checkpoint to each of the folders
            shutil.copy(
                src=os.path.join(baseModel.prefix, 'checkpoint.npz'),
                dst=os.path.join(tmp_path, 'matrix', j_folder))

        pp.run_jobs()

        # first job save dimension
        fjsi = pp.job_list[0].deltamodel._save_iter

        # proceed with assertions
        for j, j_job in enumerate(pp.job_list):
            exp_nc_path = os.path.join(
                tmp_path, 'matrix', j_folder, 'pyDeltaRCM_output.nc')
            assert os.path.isfile(exp_nc_path)

            # check all jobs have same dimensionality
            jsi = j_job.deltamodel._save_iter
            assert jsi == fjsi

            # open the file and check dimensions
            output = Dataset(exp_nc_path, 'r', allow_pickle=True)
            out_vars = output.variables.keys()
            # check that expected variables are in the file
            assert 'x' in out_vars
            assert 'y' in out_vars
            assert 'time' in out_vars
            assert 'eta' in out_vars
            # check attributes of variables

            # this is the critical check, that the dimension of the second
            # netcdf is not expanded to begin at _save_ter from the initial
            # basemodel
            assert output['time'].shape[0] <= (bmsi // 2)

    @pytest.mark.skipif(
        platform.system() == 'Windows',
        reason='OS differences regarding netCDF permissions.')
    def test_load_checkpoint_with_open_netcdf(self, tmp_path):
        """Test what happens if output netCDF file is actually open.

        This is not the same as when the netCDF file is open in another
        process. That situation raises an error for all OS.
        """
        # define a yaml with outputs (defaults will output strata)
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True,
                                      'save_eta_grids': True})
        _delta = DeltaModel(input_file=p)

        # replace eta with a random field for checkpointing success check
        _rand_field = np.random.uniform(0, 1, size=_delta.eta.shape)
        _delta.eta = _rand_field

        _delta._save_time_since_checkpoint = float("inf")
        _delta.output_checkpoint()  # force another checkpoint
        _delta.finalize()

        # paths exists
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'pyDeltaRCM_output.nc'))
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'checkpoint.npz'))
        _prefix = _delta.prefix
        _delta = []  # clear

        # open the netCDF file
        _opened = Dataset(os.path.join(_prefix, 'pyDeltaRCM_output.nc'),
                          'r+', format='NETCDF4')
        assert type(_opened) == Dataset
        assert 'eta' in _opened.variables.keys()
        # saved grid is the initial one, before random field was assigned
        assert np.all(_opened.variables['eta'][:].data != _rand_field)

        # can be resumed
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True,
                                      'resume_checkpoint': True,
                                      'save_eta_grids': True})
        _delta = DeltaModel(input_file=p)
        # force save grids/figs
        _delta.save_grids_and_figs()

        # check that fields match
        assert np.all(_delta.eta == _rand_field)
        # assert that old netCDF object is still around and hasn't changed
        assert type(_opened) == Dataset
        assert 'eta' in _opened.variables.keys()
        # grid from old netCDF is initial one, before random field was assigned
        assert np.all(_opened.variables['eta'][:].data != _rand_field)

        # clear delta
        _delta = []
        # open the new netCDF file
        _new = Dataset(os.path.join(_prefix, 'pyDeltaRCM_output.nc'),
                       'r+', format='NETCDF4')
        # first grid should be the OG one
        assert np.all(_opened['eta'][:].data == _new['eta'][0, :, :].data)
        # random field should be saved in the new netCDF file
        # some rounding/truncation happens in the netCDF so we use approx
        assert pytest.approx(_rand_field) == _new['eta'][1, :, :].data

    @pytest.mark.skipif(
        platform.system() != 'Windows',
        reason='OS differences regarding netCDF permissions.')
    def test_load_checkpoint_with_open_netcdf_win(self, tmp_path):
        """Test what happens if output netCDF file is actually open.

        This is not the same as when the netCDF file is open in another
        process. That situation raises an error for all OS.
        """
        # define a yaml with outputs (defaults will output strata)
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True,
                                      'save_eta_grids': True})
        _delta = DeltaModel(input_file=p)

        # replace eta with a random field for checkpointing success check
        _rand_field = np.random.uniform(0, 1, size=_delta.eta.shape)
        _delta.eta = _rand_field

        _delta._save_time_since_checkpoint = float("inf")
        _delta.output_checkpoint()  # force another checkpoint
        _delta.finalize()

        # paths exists
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'pyDeltaRCM_output.nc'))
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'checkpoint.npz'))
        _prefix = _delta.prefix
        _delta = []  # clear

        # open the netCDF file
        _opened = Dataset(os.path.join(_prefix, 'pyDeltaRCM_output.nc'),
                          'r+', format='NETCDF4')
        assert type(_opened) == Dataset
        assert 'eta' in _opened.variables.keys()
        # saved grid is the initial one, before random field was assigned
        assert np.all(_opened.variables['eta'][:].data != _rand_field)

        # can be resumed
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True,
                                      'resume_checkpoint': True,
                                      'save_eta_grids': True})
        # raises a permissions error on Windows
        with pytest.raises(PermissionError):
            _ = DeltaModel(input_file=p)



import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

import pyDeltaRCM


# filter out the warning raised about no netcdf being found
warnings.filterwarnings("ignore", category=UserWarning)


n = 20
cm = matplotlib.cm.get_cmap('tab10')


# init delta model
with pyDeltaRCM.shared_tools._docs_temp_directory() as output_dir:

    delta = pyDeltaRCM.DeltaModel(
        out_dir=output_dir,
        resume_checkpoint='../../_resources/checkpoint')


_shp = delta.eta.shape

# set up axis
fig, ax = plt.subplots(1, 2, figsize=(6, 3))
vmin, vmax = delta.eta.min(), delta.eta.max()

# fill in axis
ax[0].imshow(
    delta.eta, vmin=vmin, vmax=vmax, cmap='cividis')
ax[0].set_xticks([])
ax[0].set_yticks([])
ax[0].set_title('bed elevation')

ax[1].imshow(
    delta.uw, cmap='plasma')
ax[1].set_xticks([])
ax[1].set_yticks([])
ax[1].set_title('velocity')

plt.tight_layout()
plt.show()



import pytest

import numpy as np
import matplotlib.pyplot as plt

from pyDeltaRCM.model import DeltaModel
from . import utilities


class TestShowAttributeDomain:

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_domain_cell_type(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        _delta.show_attribute('cell_type')
        return plt.gcf()

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_iwalk(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        _delta.show_attribute('iwalk')
        return plt.gcf()

    def test_plot_attribute_bad_shape_1d(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        with pytest.raises(ValueError):
            _delta.show_attribute('free_surf_flag')

    def test_plot_attribute_bad_shape_3d(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # not sure if there are any 3d arrays actually, so just make a fake one
        _delta.threedeearray = np.zeros((10, 10, 2))
        with pytest.raises(ValueError):
            _delta.show_attribute('threedeearray')

    def test_plot_attribute_missing_unknown(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        with pytest.raises(AttributeError):
            _delta.show_attribute('FAKE_NAME_ATTRIBUTE')

    def test_plot_attribute_numeric_bad_type(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        with pytest.raises(TypeError):
            _delta.show_attribute('dx')
        with pytest.raises(TypeError):
            _delta.show_attribute('CTR')

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_domain_cell_type_no_grid(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        _delta.show_attribute('cell_type', grid=False)
        return plt.gcf()

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_domain_velocity(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        _delta.show_attribute('ux')
        return plt.gcf()

    def test_plot_domain_badtype(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        with pytest.raises(TypeError):
            _delta.show_attribute(42)

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_multiple_subplots(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(1, 3, figsize=(6, 4))
        _delta.show_attribute('eta', grid=False, ax=ax[1])
        _delta.show_attribute('cell_type', grid=False, ax=ax[0])
        _delta.show_attribute('ux', grid=False, ax=ax[2])
        return plt.gcf()

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_domain_withlabel(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        # This is a weak test, but it triggers coverage of the label lines.
        _delta.show_attribute('ux', label='')
        return plt.gcf()


class TestShowInd:

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_domain_cell_type_single_tuple(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        ax.imshow(_delta.cell_type, interpolation='none')
        ax.autoscale(False)
        _delta.show_ind((30, 100))
        return plt.gcf()

    def test_plot_domain_cell_type_single_badtuple(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        with pytest.raises(ValueError):
            _delta.show_ind((100, 1100, 2100))

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_domain_cell_type_list_tuple(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        ax.imshow(_delta.cell_type, interpolation='none')
        ax.autoscale(False)
        _delta.show_ind([(30, 100), (50, 100), (50, 150)])
        return plt.gcf()

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_domain_cell_type_single_index(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        ax.imshow(_delta.cell_type, interpolation='none')
        ax.autoscale(False)
        _delta.show_ind(2200)
        return plt.gcf()

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_domain_cell_type_list_index(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        ax.imshow(_delta.cell_type, interpolation='none')
        ax.autoscale(False)
        _delta.show_ind([1125, 1126, 1127, 1129, 2120])
        return plt.gcf()

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_domain_cell_type_list_mix_tuple_index(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        ax.imshow(_delta.cell_type, interpolation='none')
        ax.autoscale(False)
        _delta.show_ind([(15, 55), (15, 65), 1275, 1295, 2250])
        return plt.gcf()

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_domain_cell_type_multiple_index_calls(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        ax.imshow(_delta.cell_type, interpolation='none')
        ax.autoscale(False)
        _delta.show_ind((3, 4))
        _delta.show_ind((3, 5))
        _delta.show_ind((3, 6))
        return plt.gcf()

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_domain_cell_type_multiple_diff_args(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        ax.imshow(_delta.cell_type, interpolation='none')
        ax.autoscale(False)
        _delta.show_ind((3, 4), 'bs')
        _delta.show_ind((3, 5), 'r*')
        _delta.show_ind((3, 6), 'g^')
        return plt.gcf()

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_domain_cell_type_multiple_diff_kwargs(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        ax.imshow(_delta.cell_type, interpolation='none')
        ax.autoscale(False)
        _delta.show_ind((3, 4), marker='s', color='cyan')
        _delta.show_ind((3, 5), markersize=12)
        _delta.show_ind([(3, 4), (3, 5), (3, 6), 19, 20],
                        color='green', marker='.')
        return plt.gcf()


class TestCombinationsOfMultipleMethods:

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_plot_domain_cell_type_no_grid(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        _delta.show_attribute('cell_type', grid=False)
        _delta.show_ind([(3, 4), (3, 5), (3, 6), 19, 20],
                        color='green', marker='.')
        return plt.gcf()


class TestShowLine:

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_show_line_pts_Nx2_array(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        _arr = np.column_stack((np.arange(9), np.arange(11, 20)))
        assert _arr.shape[1] == 2

        fig, ax = plt.subplots(figsize=(5, 4))
        _delta.show_line(_arr, ax=ax)
        _delta.show_line(np.fliplr(_arr), 'r-.', ax=ax)
        return plt.gcf()

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_show_line_set_points(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        fig, ax = plt.subplots(figsize=(5, 4))
        np.random.seed(0)
        _delta.free_surf_walk_inds = np.tile(
            np.arange(100, 2900, step=_delta.eta.shape[1]),
            (12, 1))
        _shape = _delta.free_surf_walk_inds.shape
        _delta.free_surf_walk_inds += np.random.randint(-5, 5, size=_shape)

        _delta.show_line(
            _delta.free_surf_walk_inds.T, multiline=True, ax=ax)
        return plt.gcf()



import pytest
import numpy as np

from .. import utilities
from pyDeltaRCM import DeltaModel


class TestConsistentOutputsBetweenMerges:

    def test_bed_after_one_update(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'Length', 10.0)
        utilities.write_parameter_to_file(f, 'Width', 10.0)
        utilities.write_parameter_to_file(f, 'seed', 0)
        utilities.write_parameter_to_file(f, 'dx', 1.0)
        utilities.write_parameter_to_file(f, 'L0_meters', 1.0)
        utilities.write_parameter_to_file(f, 'S0', 0.0002)
        utilities.write_parameter_to_file(f, 'itermax', 1)
        utilities.write_parameter_to_file(f, 'Np_water', 10)
        utilities.write_parameter_to_file(f, 'u0', 1.0)
        utilities.write_parameter_to_file(f, 'N0_meters', 2.0)
        utilities.write_parameter_to_file(f, 'h0', 1.0)
        utilities.write_parameter_to_file(f, 'H_SL', 0.0)
        utilities.write_parameter_to_file(f, 'SLR', 0.001)
        utilities.write_parameter_to_file(f, 'Np_sed', 10)
        utilities.write_parameter_to_file(f, 'f_bedload', 0.5)
        utilities.write_parameter_to_file(f, 'C0_percent', 0.1)
        utilities.write_parameter_to_file(f, 'toggle_subsidence', False)
        utilities.write_parameter_to_file(f, 'sigma_max', 0.0)
        utilities.write_parameter_to_file(f, 'start_subsidence', 50.)
        utilities.write_parameter_to_file(f, 'save_eta_figs', False)
        utilities.write_parameter_to_file(f, 'save_stage_figs', False)
        utilities.write_parameter_to_file(f, 'save_depth_figs', False)
        utilities.write_parameter_to_file(f, 'save_discharge_figs', False)
        utilities.write_parameter_to_file(f, 'save_velocity_figs', False)
        utilities.write_parameter_to_file(f, 'save_eta_grids', False)
        utilities.write_parameter_to_file(f, 'save_stage_grids', False)
        utilities.write_parameter_to_file(f, 'save_depth_grids', False)
        utilities.write_parameter_to_file(f, 'save_discharge_grids', False)
        utilities.write_parameter_to_file(f, 'save_velocity_grids', False)
        utilities.write_parameter_to_file(f, 'save_dt', 500)
        f.close()
        _delta = DeltaModel(input_file=p)

        _delta.update()

        # slice is: _delta.eta[:5, 4]
        _exp = np.array([-1., -0.9152762, -1.0004134, -1., -1.])
        assert np.all(_delta.eta[:5, 4] == pytest.approx(_exp))

    def test_bed_after_ten_updates(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'Length', 10.0)
        utilities.write_parameter_to_file(f, 'Width', 10.0)
        utilities.write_parameter_to_file(f, 'seed', 0)
        utilities.write_parameter_to_file(f, 'dx', 1.0)
        utilities.write_parameter_to_file(f, 'L0_meters', 1.0)
        utilities.write_parameter_to_file(f, 'S0', 0.0002)
        utilities.write_parameter_to_file(f, 'itermax', 1)
        utilities.write_parameter_to_file(f, 'Np_water', 10)
        utilities.write_parameter_to_file(f, 'u0', 1.0)
        utilities.write_parameter_to_file(f, 'N0_meters', 2.0)
        utilities.write_parameter_to_file(f, 'h0', 1.0)
        utilities.write_parameter_to_file(f, 'H_SL', 0.0)
        utilities.write_parameter_to_file(f, 'SLR', 0.001)
        utilities.write_parameter_to_file(f, 'Np_sed', 10)
        utilities.write_parameter_to_file(f, 'f_bedload', 0.5)
        utilities.write_parameter_to_file(f, 'C0_percent', 0.1)
        utilities.write_parameter_to_file(f, 'toggle_subsidence', False)
        utilities.write_parameter_to_file(f, 'sigma_max', 0.0)
        utilities.write_parameter_to_file(f, 'start_subsidence', 50.)
        utilities.write_parameter_to_file(f, 'save_eta_figs', False)
        utilities.write_parameter_to_file(f, 'save_stage_figs', False)
        utilities.write_parameter_to_file(f, 'save_depth_figs', False)
        utilities.write_parameter_to_file(f, 'save_discharge_figs', False)
        utilities.write_parameter_to_file(f, 'save_velocity_figs', False)
        utilities.write_parameter_to_file(f, 'save_eta_grids', False)
        utilities.write_parameter_to_file(f, 'save_stage_grids', False)
        utilities.write_parameter_to_file(f, 'save_depth_grids', False)
        utilities.write_parameter_to_file(f, 'save_discharge_grids', False)
        utilities.write_parameter_to_file(f, 'save_velocity_grids', False)
        utilities.write_parameter_to_file(f, 'save_dt', 500)
        f.close()
        _delta = DeltaModel(input_file=p)

        for _ in range(0, 10):
            _delta.update()

        # slice is: test_DeltaModel.eta[:5, 4]
        _exp = np.array([1.7, 0.394864, -0.95006764,  -1., -1.])
        assert np.all(_delta.eta[:5, 4] == pytest.approx(_exp))

    def test_long_multi_validation(self, tmp_path):
        # IndexError on corner.

        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'seed', 42)
        utilities.write_parameter_to_file(f, 'Length', 600.)
        utilities.write_parameter_to_file(f, 'Width', 600.)
        utilities.write_parameter_to_file(f, 'dx', 5)
        utilities.write_parameter_to_file(f, 'Np_water', 10)
        utilities.write_parameter_to_file(f, 'Np_sed', 10)
        utilities.write_parameter_to_file(f, 'f_bedload', 0.05)
        f.close()
        delta = DeltaModel(input_file=p)

        for _ in range(0, 3):
            delta.update()

        # slice is: test_DeltaModel.eta[:5, 62]
        _exp1 = np.array([-4.976912, -4.979, -7.7932253, -4.9805, -2.7937498])
        assert np.all(delta.eta[:5, 62] == pytest.approx(_exp1))

        for _ in range(0, 10):
            delta.update()

        _exp2 = np.array([-4.9614887, -3.4891236, -12.195051,  -4.6706524, -2.7937498])
        assert np.all(delta.eta[:5, 62] == pytest.approx(_exp2))
        delta.finalize()


class TestModelIsReproducible:

    def test_same_result_two_models(self, tmp_path):
        """Test consistency of two models initialized from same yaml."""
        p1 = utilities.yaml_from_dict(tmp_path, 'input_1.yaml',
                                      {'out_dir': tmp_path / 'out_dir_1',
                                       'seed': 10, 'save_sandfrac_grids': True})
        p2 = utilities.yaml_from_dict(tmp_path, 'input_2.yaml',
                                      {'out_dir': tmp_path / 'out_dir_2',
                                       'seed': 10, 'save_sandfrac_grids': True})

        # create and update first model
        ModelA = DeltaModel(input_file=p1)
        ModelA.update()
        ModelA.output_netcdf.close()
        # create and update second model
        ModelB = DeltaModel(input_file=p2)
        ModelB.update()
        ModelB.output_netcdf.close()

        # fields should be the same
        assert ModelA.time == ModelB.time
        assert ModelA._time_iter == ModelB._time_iter
        assert ModelA._save_iter == ModelB._save_iter
        assert ModelA._save_time_since_data == ModelB._save_time_since_data
        assert np.all(ModelA.uw == ModelB.uw)
        assert np.all(ModelA.ux == ModelB.ux)
        assert np.all(ModelA.uy == ModelB.uy)
        assert np.all(ModelA.depth == ModelB.depth)
        assert np.all(ModelA.stage == ModelB.stage)
        assert np.all(ModelA.sand_frac == ModelB.sand_frac)
        assert np.all(ModelA.active_layer == ModelB.active_layer)


class CustomParamModel(DeltaModel):
    """Subclass for custom yaml parameters."""
    def __init__(self, input_file=None, defer_output=False, **kwargs):
        super().__init__(input_file, **kwargs)

    def hook_import_files(self):
        """Hook to define custom yaml parameters."""
        self.subclass_parameters['new_str'] = {
            'type': ['str'], 'default': 'DefaultString'
        }
        self.subclass_parameters['new_val'] = {
            'type': ['int', 'float'], 'default': 0
        }


class TestCustomParams:

    def test_custom_defaults(self, tmp_path):
        """Default subclass yaml parameters."""
        file_name = 'user_parameters.yaml'
        p = utilities.yaml_from_dict(tmp_path, file_name)
        # initialize model
        _delta = CustomParamModel(input_file=p)
        # assert that hook has been used and default params exist
        assert _delta.subclass_parameters['new_str']['default'] == \
            'DefaultString'
        assert _delta.subclass_parameters['new_val']['default'] == 0
        assert hasattr(_delta, 'subclass_parameters')
        assert _delta.new_str == 'DefaultString'
        assert _delta.new_val == 0

    def test_yaml_custom_params(self, tmp_path):
        """Specify custom params in yaml."""
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'new_str', 'Customized')
        utilities.write_parameter_to_file(f, 'new_val', 10.5)
        f.close()
        # initialize model
        _delta = CustomParamModel(input_file=p)
        # assert that hook has been used and default params exist
        assert _delta.subclass_parameters['new_str']['default'] == \
            'DefaultString'
        assert _delta.subclass_parameters['new_val']['default'] == 0
        assert hasattr(_delta, 'subclass_parameters')
        assert _delta.new_str == 'Customized'
        assert _delta.new_val == 10.5

    def test_kwargs_custom_params(self, tmp_path):
        """Try to specify one custom parameter via kwargs."""
        file_name = 'user_parameters.yaml'
        p = utilities.yaml_from_dict(tmp_path, file_name)
        # initialize model
        _delta = CustomParamModel(input_file=p, new_val=-5.25)
        # assert that hook has been used and default params exist
        assert _delta.subclass_parameters['new_str']['default'] == \
            'DefaultString'
        assert _delta.subclass_parameters['new_val']['default'] == 0
        assert hasattr(_delta, 'subclass_parameters')
        assert _delta.new_str == 'DefaultString'
        assert _delta.new_val == -5.25

    def test_invalid_custom_params(self, tmp_path):
        """Specify invalid custom param type in yaml."""
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'new_val', 'invalid_str')
        f.close()
        # initialize model
        with pytest.raises(TypeError):
            _ = CustomParamModel(input_file=p)



import matplotlib.pyplot as plt
import numpy as np


import _base


fig, ax = plt.subplots(figsize=(8, 2))
plt.subplots_adjust(top=0.7)

nflood = np.sum(_base.flood)
day = 0
for i in range(4):
    ax.axvline(x=day, color='k', ls='-')
    ax.fill_between(day + np.arange(nflood + 1), np.zeros((nflood + 1,)),
                    4 * np.ones((nflood + 1,)),
                    alpha=0.4, edgecolor='none')
    ax.plot(day + np.arange(nflood + 1), 2.5 * np.ones((nflood + 1),))
    day += nflood

ax.axvline(x=day, color='k', ls='-')
ax.set_ylim(0.9, 3.1)
ax.set_xlim(0 - 10, 365 * 4 + 10)
ax.set_yticks([])
ax.set_xlabel(r'time $\rightarrow$')
ax.set_yticks([1.05, 2.5])
ax.set_yticklabels(['base flow', 'bankfull flow'])

plt.show()



# unit tests for sed_tools.py

import numpy as np

import pytest
import unittest.mock as mock

from pyDeltaRCM.model import DeltaModel
from . import utilities


class TestSedimentRoute:

    def test_route_sediment(self, tmp_path):
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # mock top-level methods
        _delta.log_info = mock.MagicMock()
        _delta.init_sediment_iteration = mock.MagicMock()
        _delta.route_all_sand_parcels = mock.MagicMock()
        _delta.topo_diffusion = mock.MagicMock()
        _delta.route_all_mud_parcels = mock.MagicMock()

        # run the method
        _delta.route_sediment()

        # methods called
        assert (_delta.log_info.call_count == 4)
        assert (_delta.init_sediment_iteration.called is True)
        assert (_delta.route_all_sand_parcels.called is True)
        assert (_delta.topo_diffusion.called is True)
        assert (_delta.route_all_mud_parcels.called is True)

    def test_sed_route_deprecated(self, tmp_path):
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # mock top-level methods
        _delta.logger = mock.MagicMock()
        _delta.route_sediment = mock.MagicMock()

        # check warning raised
        with pytest.warns(UserWarning):
            _delta.sed_route()

        # and logged
        assert (_delta.logger.warning.called is True)


class TestInitSedimentIteration:

    def test_fields_cleared(self, tmp_path):
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # alter field for initial values going into function
        _delta.qs += np.random.uniform(0, 1, size=_delta.eta.shape)
        _delta.Vp_dep_sand += np.random.uniform(0, 1, size=_delta.eta.shape)
        _delta.Vp_dep_mud += np.random.uniform(0, 1, size=_delta.eta.shape)

        assert not np.all(_delta.qs == 0)  # field has info

        # call the method
        _delta.init_sediment_iteration()

        # assertions
        assert np.all(_delta.pad_depth[1:-1, 1:-1] == _delta.depth)
        assert np.all(_delta.qs == 0)      # field is cleared
        assert np.all(_delta.Vp_dep_sand == 0)
        assert np.all(_delta.Vp_dep_mud == 0)


class TestRouteAllSandParcels:

    def test_route_sand_parcels(self, tmp_path):
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'Np_sed': 1000,
                                      'f_bedload': 0.6})
        _delta = DeltaModel(input_file=p)

        # mock top-level methods / objects
        _delta.log_info = mock.MagicMock()
        _delta._sr = mock.MagicMock()

        # mock the shared tools start indices
        def _patched_starts(inlet, inlet_weights, num_starts):
            return np.random.randint(0, 5, size=(num_starts,))

        patcher = mock.patch(
            'pyDeltaRCM.shared_tools.get_start_indices',
            new=_patched_starts)
        patcher.start()

        # run the method
        _delta.route_all_sand_parcels()

        # methods called
        assert (_delta._sr.run.call_count == 1)
        assert (_delta.log_info.call_count == 3)

        # stop the patch
        patcher.stop()


class TestRouteAllMudParcels:

    def test_route_mud_parcels(self, tmp_path):
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'Np_sed': 1000,
                                      'f_bedload': 0.6})
        _delta = DeltaModel(input_file=p)

        # mock top-level methods / objects
        _delta.log_info = mock.MagicMock()
        _delta._mr = mock.MagicMock()

        # mock the shared tools start indices
        def _patched_starts(inlet, inlet_weights, num_starts):
            return np.random.randint(0, 5, size=(num_starts,))

        patcher = mock.patch(
            'pyDeltaRCM.shared_tools.get_start_indices',
            new=_patched_starts)
        patcher.start()

        # run the method
        _delta.route_all_mud_parcels()

        # methods called
        assert (_delta._mr.run.call_count == 1)
        assert (_delta.log_info.call_count == 3)

        # stop the patch
        patcher.stop()
